#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from xml.dom import Node

__all__ = ('Event', 'Node', 'NodeList', 'HTMLCollection')


class Event:
    def __init__(self, type: str, **kwargs):
        self.type = type.lower()
        self.__dict__.update(kwargs)

    def stopPrapagation(self):
        raise NotImplementedError


class NodeList:
    def __init__(self, ref: list):
        self.ref = ref

    def __getitem__(self, index: int) -> Node:
        return self.ref[index]

    def __len__(self) -> int:
        return len(self.ref)

    def __contains__(self, other: Node) -> bool:
        return other in self.ref

    def __iter__(self) -> Node:
        for n in self.ref:
            yield n

    @property
    def length(self) -> int:
        return len(self)

    def item(self, index: int) -> Node:
        '''Return item with the index. If the index is negative number or out
        of the list, return None.
        '''
        if not isinstance(index, int):
            raise TypeError(
                'Indeces must be integer, not {}'.format(type(index)))
        return self.ref[index] if 0 <= index < self.length else None

    def index(self, node: Node) -> int:
        '''Get index of the node.'''
        return self.ref.index(node)


class HTMLCollection(NodeList):
    def namedItem(self, name: str) -> Node:
        for n in self.ref:
            if n.getAttribute('id') == name:
                return n
        for n in self.ref:
            if n.getAttribute('name') == name:
                return n



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
This module defines options for wdom and wraps ``tornado.options``.
Do not use ``tornado.options`` directly.
"""

import sys
import logging
from argparse import ArgumentParser, Namespace
import re
from typing import Union

from tornado.log import LogFormatter

__all__ = [
    'parser',
    'config',
    'root_logger',
    'set_loglevel',
    'parse_command_line',
]

# Setup root logger
root_logger = logging.getLogger('wdom')  # Root logger
_log_handler = logging.StreamHandler()
fmt = '%(color)s[%(levelname)1.1s:%(name)s]%(end_color)s '
fmt += '%(message)s'
formatter = LogFormatter(fmt=fmt)
_log_handler.setFormatter(formatter)
root_logger.addHandler(_log_handler)
root_logger.propagate = False

# local logger
logger = logging.getLogger(__name__)

# setup argument parser
config = Namespace()
parser = ArgumentParser(prog='WDOM', argument_default=None)
parser.add_argument(
    '--logging', choices=['debug', 'info', 'warn', 'error'],
    help='Set the log level (dafualt: `info`).',
)
parser.add_argument(
    '--debug', default=False, action='store_const', const=True,
    help='Enable debug mode (dafualt: False).'
    ' Debug mode enables `--autoreload` and set the default log level `debug`.'
)
parser.add_argument(
    '--address', default='localhost',
    help='Address to run server (default: `localhost`).'
)
parser.add_argument(
    '--port', default=8888, type=int,
    help='Port to run server (defualt: 8888). If 0, use arbitrary free port.',
)
parser.add_argument(
    '--autoreload', default=False, action='store_const', const=True,
    help='Watch files and restart when any files changed (default: False).',
)
parser.add_argument(
    '--theme', default=None, type=str,
    help='Choose theme name to use with wdom.themes module.'
    ' By default (None) or unavailable name, use `wdom.tag`.'
)
parser.add_argument(
    '--auto-shutdown', default=False, action='store_const', const=True,
    help='Terminate server process when all connections (browser tabs) closed'
    ' (default: False).',
)
parser.add_argument(
    '--shutdown-wait', default=1.0, type=float,
    help='Seconds to wait until shutdown after all connections closed'
    ' when --auto-shutdown is enabled (default: 1.0 [sec]).',
)
parser.add_argument(
    '--open-browser', default=False, action='store_const', const=True,
    help='Open browser automatically (default: False).',
)
parser.add_argument(
    '--browser', default=None, help='Browser name to open.'
    ' Only affects when used with --open-browser option.'
    ' Available values are keys of `webbrowser._browsers`.'
    ' When not specified or specified invalid value, open system\'s'
    ' default browser (default: None).',
)


def level_to_int(level: Union[str, int]):
    if isinstance(level, int):
        if logging.NOTSET <= level <= logging.FATAL:
            return level
        else:
            raise ValueError('Log level must be 0 <= level <= 50,'
                             'but gat: {}'.format(level))
    elif isinstance(level, str):
        try:
            return getattr(logging, level.upper())
        except AttributeError:
            raise ValueError('Invalid log level: {}'.format(level))
    else:
        raise TypeError(
            'Log level must be int (0 ~ 50) or string,'
            'but gat type: {}'.format(type(level)))


def set_loglevel(level=None):
    if level is not None:
        lv = level_to_int(level)
    elif config.logging:
        lv = level_to_int(config.logging)
    elif config.debug:
        lv = logging.DEBUG
    else:
        lv = logging.INFO
    root_logger.setLevel(lv)
    _log_handler.setLevel(lv)


def parse_command_line():
    '''Parse command line options and set options in ``tornado.options``.'''
    import tornado.options
    _, unkown_args = parser.parse_known_args(namespace=config)
    set_loglevel()
    if unkown_args and not re.search(r'py\.test[-.0-9]*$', sys.argv[0]):
        # warn when get unknown argument
        # if run in test, skip warning since test runner adds some arguments
        logger.warning('Unknown Arguments: {}'.format(unkown_args))
        parser.print_help()
    for k, v in vars(config).items():
        if k.startswith('log'):
            tornado.options.options.__setattr__(k, v)
    return config


parse_command_line()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import unittest

from wdom import testing
from wdom.document import get_document


class TestInitialize(unittest.TestCase):
    def test_initialize(self):
        from wdom.server import _tornado
        old_doc = get_document()
        old_app_tornado = _tornado.get_app()
        testing.reset()
        self.assertIsNot(old_doc, get_document())
        self.assertIsNot(old_app_tornado, _tornado.get_app())
        try:
            from wdom.server import _aiohttp
            old_app_aio = _aiohttp.get_app()
            testing.reset()
            self.assertIsNot(old_app_aio, _aiohttp.get_app())
        except ImportError:
            pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
from collections import Iterable
from typing import Tuple, Union

from wdom.node import Node
from wdom.element import DOMTokenList, ElementMeta
from wdom.element import (
    HTMLAnchorElement,
    HTMLButtonElement,
    HTMLFormElement,
    HTMLIFrameElement,
    HTMLInputElement,
    HTMLLabelElement,
    HTMLOptGroupElement,
    HTMLOptionElement,
    HTMLScriptElement,
    HTMLSelectElement,
    HTMLStyleElement,
    HTMLTextAreaElement,
)
from wdom.web_node import WebElement

logger = logging.getLogger(__name__)


class HTMLElement(WebElement):
    pass


class TagBaseMeta(ElementMeta):
    '''Meta class to set default class variable of HtmlDom'''
    def __prepare__(name, bases, **kwargs) -> dict:
        return {'inherit_class': True}


class Tag(HTMLElement, metaclass=TagBaseMeta):
    '''Base class for html tags. ``HTMLElement`` requires to specify tag name
    when instanciate it, but this class and sublasses have default tag name and
    not need to specify it for each thier instances.

    Additionally, this class provides shortcut properties to handle some
    special attributes (class, type, is).
    '''
    #: Tag name used for this node.
    tag = 'tag'
    #: str and list of strs are acceptale.
    class_ = ''
    #: Inherit classes defined in super class or not.
    #: By default, this variable is True.
    inherit_class = True
    #: use for <input> tag's type
    type_ = ''
    #: custom element which extends built-in tag (like <table is="your-tag">)
    is_ = ''

    def __init__(self, *args, attrs=None, **kwargs):
        if attrs:
            kwargs.update(attrs)
        if self.type_ and 'type' not in kwargs:
            kwargs['type'] = self.type_
        if self.is_ and 'is' not in kwargs and 'is_' not in kwargs:
            kwargs['is'] = self.is_
        super().__init__(self.tag, **kwargs)
        self.append(*args)

    @classmethod
    def get_class_list(cls) -> DOMTokenList:
        '''Return class-level class list, including all super class's.
        '''
        l = []
        l.append(DOMTokenList(cls, cls.class_))
        if cls.inherit_class:
            for base_cls in cls.__bases__:
                if issubclass(base_cls, Tag):
                    l.append(base_cls.get_class_list())
        # Reverse order so that parent's class comes to front
        l.reverse()
        return DOMTokenList(cls, l)

    def __getitem__(self, attr: Union[str, int]) -> Union[Node, str]:
        if isinstance(attr, int):
            return self.childNodes[attr]
        else:
            return self.getAttribute(attr)

    def __setitem__(self, attr: str, val):
        self.setAttribute(attr, val)

    def __delitem__(self, attr: str):
        self.removeAttribute(attr)

    def __copy__(self) -> HTMLElement:
        clone = type(self)()
        for attr in self.attributes:
            clone.setAttribute(attr, self.getAttribute(attr))
        for c in self.classList:
            clone.addClass(c)
        return clone

    def getAttribute(self, attr: str) -> str:
        if attr == 'class':
            cls = self.get_class_list()
            cls._append(self.classList)
            if cls:
                return cls.toString()
            else:
                return None
        else:
            return super().getAttribute(attr)

    def addClass(self, *classes: Tuple[str]):
        self.classList.add(*classes)

    def hasClass(self, class_: str) -> bool:
        return class_ in self.classList

    def hasClasses(self) -> bool:
        return len(self.classList) > 0

    def removeClass(self, *classes: Tuple[str]):
        _remove_cl = []
        for class_ in classes:
            if class_ not in self.classList:
                if class_ in self.__class__.get_class_list():
                    logger.warning(
                        'tried to remove class-level class: '
                        '{}'.format(class_)
                    )
                else:
                    logger.warning(
                        'tried to remove non-existing class: {}'.format(class_)
                    )
            else:
                _remove_cl.append(class_)
        self.classList.remove(*_remove_cl)

    def show(self):
        self.hidden = False

    def hide(self):
        self.hidden = True

    @property
    def type(self) -> str:
        return self.getAttribute('type') or self.type_

    @type.setter
    def type(self, val: str):
        self.setAttribute('type', val)


class NestedTag(Tag):
    #: Inner nested tag class
    inner_tag_class = None

    def __init__(self, *args, **kwargs):
        self._inner_element = None
        super().__init__(**kwargs)
        if self.inner_tag_class:
            self._inner_element = self.inner_tag_class()
            super().appendChild(self._inner_element)
        self.append(*args)

    def appendChild(self, child) -> Node:
        if self._inner_element:
            return self._inner_element.appendChild(child)
        else:
            return super().appendChild(child)

    def insertBefore(self, child: Node, ref_node: Node) -> Node:
        if self._inner_element:
            return self._inner_element.insertBefore(child, ref_node)
        else:
            return super().insertBefore(child, ref_node)

    def removeChild(self, child: Node) -> Node:
        if self._inner_element:
            return self._inner_element.removeChild(child)
        else:
            return super().removeChild(child)

    def replaceChild(self, new_child: Node, old_child: Node) -> Node:
        if self._inner_element:
            return self._inner_element.replaceChild(new_child, old_child)
        else:
            return super().replaceChild(new_child, old_child)

    @property
    def childNodes(self):
        if self._inner_element:
            return self._inner_element.childNodes
        else:
            return super().childNodes

    def empty(self):
        if self._inner_element:
            self._inner_element.empty()
        else:
            super().empty()

    @property
    def textContent(self) -> str:
        return super().textContent

    @textContent.setter
    def textContent(self, text: str):
        if self._inner_element:
            self._inner_element.textContent = text
        else:
            # Need a trick to call property of super-class
            Tag.textContent.fset(self, text)

    @property
    def html(self) -> str:
        if self._inner_element:
            return self.start_tag + self._inner_element.html + self.end_tag
        else:
            return Tag.html.fget(self)

    @property
    def innerHTML(self) -> str:
        if self._inner_element:
            return self._inner_element.innerHTML
        else:
            return Tag.innerHTML.fget(self)

    @innerHTML.setter
    def innerHTML(self, html: str):
        if self._inner_element:
            self._inner_element.innerHTML = html
        else:
            Tag.innerHTML.fset(self, html)


def NewTagClass(class_name: str, tag: str=None, bases: Tuple[type]=(Tag, ),
                **kwargs) -> type:
    '''Generate and return new ``Tag`` class. If ``tag`` is empty, lower case
    of ``class_name`` is used for a tag name of the new class. ``bases`` should
    be a tuple of base classes. If it is empty, use ``Tag`` class for a base
    class. Other keyword arguments are used for class variables of the new
    class.

    Example::

        MyButton = NewTagClass('MyButton', 'button', (Button,),
                               class_='btn', is_='my-button')
        my_button = MyButton('Click!')
        print(my_button.html)

        >>> <button class="btn" id="111111111">Click!</button>
    '''
    if tag is None:
        tag = class_name.lower()
    if type(bases) is not tuple:
        if isinstance(bases, Iterable):
            bases = tuple(bases)
        elif isinstance(bases, type):
            bases = (bases, )
        else:
            TypeError('Invalid base class: {}'.format(str(bases)))
    kwargs['tag'] = tag
    cls = type(class_name, bases, kwargs)
    return cls


class Input(Tag, HTMLInputElement):
    '''Base class for ``<input>`` element.
    '''
    tag = 'input'
    #: type attribute; text, button, checkbox, or radio... and so on.
    type_ = ''

    def __init__(self, *args, **kwargs) -> None:
        if self.type_ and 'type' not in kwargs:
            kwargs['type'] = self.type_
        super().__init__(*args, **kwargs)


class Textarea(Tag, HTMLTextAreaElement):
    '''Base class for ``<textarea>`` element.'''
    tag = 'textarea'

    @property
    def value(self) -> str:
        '''Get input value of this node. This value is used as a default value
        of this element.
        '''
        return self.textContent

    @value.setter
    def value(self, value: str) -> None:
        self.textContent = value


class Script(Tag, HTMLScriptElement):
    tag = 'script'

    def __init__(self, *args, type='text/javascript', **kwargs):
        super().__init__(*args, type=type, **kwargs)


Html = NewTagClass('Html')
Body = NewTagClass('Body')
Meta = NewTagClass('Meta')
Head = NewTagClass('Head')
Link = NewTagClass('Link')
Title = NewTagClass('Title')
Style = NewTagClass('Style', 'style', (Tag, HTMLStyleElement))
Iframe = NewTagClass('Iframe', 'iframe', (Tag, HTMLIFrameElement))

Div = NewTagClass('Div')
Span = NewTagClass('Span')

# Typography
H1 = NewTagClass('H1')
H2 = NewTagClass('H2')
H3 = NewTagClass('H3')
H4 = NewTagClass('H4')
H5 = NewTagClass('H5')
H6 = NewTagClass('H6')

P = NewTagClass('P')
A = NewTagClass('A', 'a', (Tag, HTMLAnchorElement))
Strong = NewTagClass('Strong')
Em = NewTagClass('Em')
U = NewTagClass('U')
Br = NewTagClass('Br')
Hr = NewTagClass('Hr')

Cite = NewTagClass('Cite')
Code = NewTagClass('Code')
Pre = NewTagClass('Pre')

Img = NewTagClass('Img')

# table tags
Table = NewTagClass('Table')
Thead = NewTagClass('Thead')
Tbody = NewTagClass('Tbody')
Tfoot = NewTagClass('Tfoot')
Th = NewTagClass('Th')
Tr = NewTagClass('Tr')
Td = NewTagClass('Td')

# List tags
Ol = NewTagClass('Ol')
Ul = NewTagClass('Ul')
Li = NewTagClass('Li')

# Definition-list tags
Dl = NewTagClass('Dl')
Dt = NewTagClass('Dt')
Dd = NewTagClass('Dd')

# Form controls
Form = NewTagClass('Form', 'form', (Tag, HTMLFormElement))
Button = NewTagClass('Button', 'button', (Tag, HTMLButtonElement))
Label = NewTagClass('Label', 'label', (Tag, HTMLLabelElement))
TextInput = NewTagClass('TextInput', 'input', Input, type_='text')
CheckBox = NewTagClass('CheckBox', 'input', Input, type_='checkbox')
RadioButton = NewTagClass('RadioButton', 'input', Input, type_='radio')
Select = NewTagClass('Select', 'select', (NestedTag, HTMLSelectElement))
Optgroup = NewTagClass('OptGroup', 'optgroup', (Tag, HTMLOptGroupElement))
Option = NewTagClass('Option', 'option', (Tag, HTMLOptionElement))

# Building blocks
Container = NewTagClass('Container', 'div', Div, is_='container')
Wrapper = NewTagClass('Wrapper', 'div', Div, is_='wrapper')
Row = NewTagClass('Row', 'div', Div, is_='row')
FormGroup = NewTagClass('FormGroup', 'div', Div, is_='formgroup')
Col = NewTagClass('Col', 'div', Div, is_='col')
Col1 = NewTagClass('Col1', 'div', Div, is_='col1')
Col2 = NewTagClass('Col2', 'div', Div, is_='col2')
Col3 = NewTagClass('Col3', 'div', Div, is_='col3')
Col4 = NewTagClass('Col4', 'div', Div, is_='col4')
Col5 = NewTagClass('Col5', 'div', Div, is_='col5')
Col6 = NewTagClass('Col6', 'div', Div, is_='col6')
Col7 = NewTagClass('Col7', 'div', Div, is_='col7')
Col8 = NewTagClass('Col8', 'div', Div, is_='col8')
Col9 = NewTagClass('Col9', 'div', Div, is_='col9')
Col10 = NewTagClass('Col10', 'div', Div, is_='col10')
Col11 = NewTagClass('Col11', 'div', Div, is_='col11')
Col12 = NewTagClass('Col12', 'div', Div, is_='col12')

# Some css updates
DefaultButton = NewTagClass('DefaultButton', 'button', Button, is_='default-button')  # noqa
PrimaryButton = NewTagClass('PrimaryButton', 'button', Button, is_='primary-button')  # noqa
SecondaryButton = NewTagClass('SecondaryButton', 'button', Button, is_='secondary-button')  # noqa
SuccessButton = NewTagClass('SuccessButton', 'button', Button, is_='success-button')  # noqa
InfoButton = NewTagClass('InfoButton', 'button', Button, is_='info-button')  # noqa
WarningButton = NewTagClass('WarningButton', 'button', Button, is_='warning-button')  # noqa
DangerButton = NewTagClass('DangerButton', 'button', Button, is_='danger-button')  # noqa
ErrorButton = NewTagClass('ErrorButton', 'button', Button, is_='error-button')  # noqa
LinkButton = NewTagClass('LinkButton', 'button', Button, is_='link-button')  # noqa

# css/js/headers
css_files = []
js_files = []
headers = []

default_classes = (
    Input,
    Textarea,
    Button,
    Script,
    Html,
    Body,
    Meta,
    Head,
    Link,
    Title,
    Style,
    Div,
    Span,
    H1,
    H2,
    H3,
    H4,
    H5,
    H6,
    P,
    A,
    Strong,
    Em,
    U,
    Br,
    Hr,
    Cite,
    Code,
    Pre,
    Img,
    Table,
    Thead,
    Tbody,
    Tfoot,
    Th,
    Tr,
    Td,
    Ol,
    Ul,
    Li,
    Dl,
    Dt,
    Dd,
    Form,
    Label,
    Option,
    Select,
)

extended_classes = [
    Container,
    Wrapper,
    Row,
    FormGroup,
    Col,
    Col1,
    Col2,
    Col3,
    Col4,
    Col5,
    Col6,
    Col7,
    Col8,
    Col9,
    Col10,
    Col11,
    Col12,
    DefaultButton,
    PrimaryButton,
    SecondaryButton,
    SuccessButton,
    InfoButton,
    WarningButton,
    DangerButton,
    ErrorButton,
    LinkButton,
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from unittest import TestCase
from wdom.node import Node, ParentNode, NonDocumentTypeChildNode, ChildNode
from wdom.node import Text, DocumentType, DocumentFragment
from wdom.node import RawHtml, Comment, CharacterData
from wdom.element import Element


class TestNode(TestCase):
    def setUp(self):
        self.node = Node()
        self.c1 = Node()
        self.c2 = Node()
        self.c3 = Node()

    def test_attributes(self):
        self.assertFalse(self.node.hasAttributes())

    def test_parent(self):
        self.assertIsNone(self.node.parentNode)
        self.assertIsNone(self.c1.parentNode)
        self.node.appendChild(self.c1)
        self.assertIs(self.node, self.c1.parentNode)

        self.node.removeChild(self.c1)
        self.assertIsNone(self.c1.parentNode)

    def test_parent_init(self):
        node = Node(parent=self.node)
        self.assertTrue(self.node.hasChildNodes())
        self.assertIs(self.node, node.parentNode)
        self.assertFalse(node.hasChildNodes())

    def test_addremove_child(self):
        self.assertFalse(self.node.hasChildNodes())
        self.assertEqual(len(self.node), 0)
        self.assertEqual(self.node.length, 0)

        appended_child1 = self.node.appendChild(self.c1)
        self.assertIs(appended_child1, self.c1)
        self.assertTrue(self.node.hasChildNodes())
        self.assertEqual(len(self.node), 1)
        self.assertEqual(self.node.length, 1)

        appended_child2 = self.node.appendChild(self.c2)
        self.assertIs(appended_child2, self.c2)
        self.assertEqual(len(self.node), 2)
        self.assertEqual(self.node.length, 2)

        removed_child2 = self.node.removeChild(self.c2)
        self.assertIs(removed_child2, self.c2)
        self.assertEqual(len(self.node), 1)
        self.assertEqual(self.node.length, 1)
        self.assertIsNone(self.c2.parentNode)

        removed_child1 = self.node.removeChild(self.c1)
        self.assertIs(removed_child1, self.c1)
        self.assertFalse(self.node.hasChildNodes())
        self.assertEqual(len(self.node), 0)
        self.assertEqual(self.node.length, 0)

        with self.assertRaises(ValueError):
            self.node.removeChild(self.c1)

    def test_empty(self):
        self.node.appendChild(self.c1)
        self.assertTrue(self.node.hasChildNodes())
        self.node.empty()
        self.assertFalse(self.node.hasChildNodes())

        self.node.appendChild(self.c1)
        self.node.appendChild(self.c2)
        self.c1.appendChild(self.c3)
        self.assertTrue(self.node.hasChildNodes())
        self.node.empty()
        self.assertFalse(self.node.hasChildNodes())
        self.assertTrue(self.c1.hasChildNodes())
        self.assertIsNone(self.c1.parentNode)

    def test_insert_before(self):
        self.node.appendChild(self.c1)
        self.node.appendChild(self.c2)
        inserted_node3 = self.node.insertBefore(self.c3, self.c2)

        self.assertIs(inserted_node3, self.c3)
        self.assertIs(self.c3.parentNode, self.node)
        self.assertIs(self.node.childNodes[0], self.c1)
        self.assertIs(self.node.childNodes[1], self.c3)
        self.assertIs(self.node.childNodes[2], self.c2)

    def test_insert_first(self):
        self.node.appendChild(self.c1)
        self.node.appendChild(self.c2)
        inserted_node3 = self.node.insertBefore(self.c3, self.c1)

        self.assertIs(inserted_node3, self.c3)
        self.assertIs(self.c3.parentNode, self.node)
        self.assertIs(self.node.childNodes[0], self.c3)
        self.assertIs(self.node.childNodes[1], self.c1)
        self.assertIs(self.node.childNodes[2], self.c2)

    def test_replace_child(self):
        self.node.appendChild(self.c1)
        self.assertTrue(self.c1 in self.node)
        self.assertFalse(self.c2 in self.node)

        replaced_node = self.node.replaceChild(self.c2, self.c1)
        self.assertIs(replaced_node, self.c1)
        self.assertFalse(self.c1 in self.node)
        self.assertTrue(self.c2 in self.node)
        self.assertIsNone(self.c1.parentNode)
        self.assertIs(self.c2.parentNode, self.node)

    def test_first_last_child(self):
        self.assertIsNone(self.node.firstChild)
        self.assertIsNone(self.node.lastChild)

        self.node.appendChild(self.c1)
        self.assertIs(self.node.firstChild, self.c1)
        self.assertIs(self.node.lastChild, self.c1)

        self.node.appendChild(self.c2)
        self.assertIs(self.node.firstChild, self.c1)
        self.assertIs(self.node.lastChild, self.c2)

    def test_siblings(self):
        self.assertIsNone(self.node.previousSibling)
        self.assertIsNone(self.node.nextSibling)

        self.node.appendChild(self.c1)
        self.assertIsNone(self.c1.previousSibling)
        self.assertIsNone(self.c1.nextSibling)

        self.node.appendChild(self.c2)
        self.assertIsNone(self.c1.previousSibling)
        self.assertIs(self.c1.nextSibling, self.c2)
        self.assertIs(self.c2.previousSibling, self.c1)
        self.assertIsNone(self.c2.nextSibling)

    def _test_shallow_copy(self, clone):
        self.assertTrue(self.node.hasChildNodes())
        self.assertFalse(clone.hasChildNodes())
        self.assertEqual(len(clone), 0)

        clone.appendChild(self.c2)
        self.assertTrue(clone.hasChildNodes())
        self.assertIn(self.c2, clone)
        self.assertNotIn(self.c2, self.node)

    def test_copy(self):
        from copy import copy
        self.node.appendChild(self.c1)
        clone = copy(self.node)
        self._test_shallow_copy(clone)

    def test_clone_node_sharrow(self):
        self.node.appendChild(self.c1)
        clone = self.node.cloneNode()
        self._test_shallow_copy(clone)

        clone2 = self.node.cloneNode(deep=False)
        self._test_shallow_copy(clone2)

    def _test_deep_copy(self, clone):
        self.assertTrue(clone.hasChildNodes())
        self.assertEqual(len(clone), 1)
        self.assertIn(self.c1, self.node)
        self.assertNotIn(self.c1, clone)

        clone.appendChild(self.c2)
        self.assertEqual(len(clone), 2)
        self.assertEqual(len(self.node), 1)

    def test_deepcopy(self):
        from copy import deepcopy
        self.node.appendChild(self.c1)
        clone = deepcopy(self.node)
        self._test_deep_copy(clone)

    def test_clone_node_deep(self):
        self.node.appendChild(self.c1)
        clone = self.node.cloneNode(deep=True)
        self._test_deep_copy(clone)

    def test_owner_document(self):
        self.assertIsNone(self.node.ownerDocument)

    def test_text_content(self):
        self.assertEqual(self.node.textContent, '')
        self.node.textContent = 'a'
        self.assertEqual(self.node.textContent, 'a')
        self.node.textContent = 'b'
        self.assertEqual(self.node.textContent, 'b')

        self.node.appendChild(self.c1)
        self.c1.textContent = 'c1'
        self.assertEqual(self.node.textContent, 'bc1')
        self.node.textContent = 'd'
        self.assertEqual(self.node.textContent, 'd')
        self.assertIsNone(self.c1.parentNode)
        self.assertEqual(self.c1.textContent, 'c1')

    def test_index(self):
        self.node.appendChild(self.c1)
        self.node.appendChild(self.c2)
        self.node.appendChild(self.c3)
        self.assertEqual(self.node.index(self.c1), 0)
        self.assertEqual(self.node.index(self.c2), 1)
        self.assertEqual(self.node.index(self.c3), 2)
        self.assertEqual(self.node.childNodes.index(self.c1), 0)
        self.assertEqual(self.node.childNodes.index(self.c2), 1)
        self.assertEqual(self.node.childNodes.index(self.c3), 2)


class P(Node, ParentNode):
    pass


class NDTC(Node, NonDocumentTypeChildNode):
    pass


class C(Node, ChildNode):
    pass


def is_equal_nodes(p, nodes2):
    nodes1 = p.childNodes
    assert len(nodes1) == len(nodes2)
    for n1, n2 in zip(nodes1, nodes2):
        if isinstance(n1, Text):
            if isinstance(n2, Text):
                assert n1.textContent == n2.textContent
            elif isinstance(n2, str):
                assert n1.textContent == n2
            else:
                raise AssertionError(n1, n2)
        elif isinstance(n1, Node):
            assert n1 is n2
        else:
            raise AssertionError(n1, n2)


class TestParentNode(TestCase):
    def setUp(self):
        self.p = P()
        self.c1 = C()
        self.c2 = C()
        self.c3 = C()
        self.c4 = C()

    def test_children(self):
        self.p.appendChild(Text('a'))
        self.p.appendChild(Comment('b'))
        self.assertEqual(self.p.children.length, 0)
        elm1 = Element('c1')
        self.p.appendChild(elm1)
        self.p.appendChild(Text('d'))
        self.assertEqual(self.p.children.length, 1)
        self.assertIs(self.p.firstElementChild, elm1)
        self.assertIs(self.p.lastElementChild, elm1)
        elm2 = Element('c2')
        self.p.appendChild(elm2)
        self.p.appendChild(Text('e'))
        self.assertIs(self.p.firstElementChild, elm1)
        self.assertIs(self.p.lastElementChild, elm2)

    def test_append(self):
        self.assertFalse(self.p.hasChildNodes())
        self.p.append(self.c1)
        self.assertTrue(self.p.hasChildNodes())
        self.assertEqual(self.p.length, 1)
        is_equal_nodes(self.p, [self.c1])

        self.p.append(self.c2)
        self.assertEqual(self.p.length, 2)
        self.assertIs(self.p.firstChild, self.c1)
        self.assertIs(self.p.lastChild, self.c2)
        is_equal_nodes(self.p, [self.c1, self.c2])

    def test_append_multi(self):
        self.p.append(self.c1, self.c2)
        self.assertTrue(self.p.hasChildNodes())
        self.assertEqual(self.p.length, 2)
        is_equal_nodes(self.p, [self.c1, self.c2])

        self.p.append(self.c3, self.c4)
        self.assertEqual(self.p.length, 4)
        is_equal_nodes(self.p, [self.c1, self.c2, self.c3, self.c4])

    def test_append_text(self):
        self.p.append('a')
        self.assertTrue(self.p.hasChildNodes())
        self.assertEqual(self.p.length, 1)
        self.assertTrue(isinstance(self.p.firstChild, Text))
        self.assertEqual(self.p.textContent, 'a')
        is_equal_nodes(self.p, ['a'])

        self.p.append('b')
        self.assertEqual(self.p.length, 2)
        self.assertTrue(isinstance(self.p.firstChild, Text))
        self.assertTrue(isinstance(self.p.lastChild, Text))
        self.assertEqual(self.p.textContent, 'ab')
        is_equal_nodes(self.p, ['a', 'b'])

        self.p.append('c', 'd')
        self.assertEqual(self.p.length, 4)
        self.assertTrue(isinstance(self.p.firstChild, Text))
        self.assertTrue(isinstance(self.p.lastChild, Text))
        self.assertEqual(self.p.textContent, 'abcd')
        is_equal_nodes(self.p, ['a', 'b', 'c', 'd'])

    def test_append_mixed_text_node(self):
        self.p.append(self.c1, 'a', self.c2)
        self.assertTrue(self.p.hasChildNodes())
        self.assertEqual(self.p.length, 3)
        self.assertIs(self.p.firstChild, self.c1)
        self.assertIs(self.p.lastChild, self.c2)
        is_equal_nodes(self.p, [self.c1, 'a', self.c2])

        self.p.append(self.c3, 'b', self.c4)
        is_equal_nodes(self.p, [self.c1, 'a', self.c2, self.c3, 'b', self.c4])

    def test_prepend(self):
        self.p.prepend(self.c1)
        self.assertTrue(self.p.hasChildNodes())
        self.assertEqual(self.p.length, 1)
        is_equal_nodes(self.p, [self.c1])
        self.p.prepend(self.c2)
        is_equal_nodes(self.p, [self.c2, self.c1])

    def test_prepend_multi(self):
        self.p.append(self.c1)
        self.p.prepend(self.c2, self.c3)
        self.assertEqual(self.p.length, 3)
        is_equal_nodes(self.p, [self.c2, self.c3, self.c1])

    def test_prepend_text(self):
        self.p.prepend('a')
        self.assertEqual(self.p.length, 1)
        self.assertTrue(isinstance(self.p.firstChild, Text))
        self.assertEqual(self.p.textContent, 'a')
        is_equal_nodes(self.p, ['a'])

        self.p.prepend('b', 'c')
        self.assertEqual(self.p.length, 3)
        self.assertTrue(isinstance(self.p.firstChild, Text))
        self.assertTrue(isinstance(self.p.lastChild, Text))
        self.assertEqual(self.p.textContent, 'bca')
        is_equal_nodes(self.p, ['b', 'c', 'a'])

    def test_prepend_mixed_text_node(self):
        self.p.prepend(self.c1, 'a', self.c2)
        self.assertTrue(self.p.hasChildNodes())
        self.assertEqual(self.p.length, 3)
        self.assertIs(self.p.firstChild, self.c1)
        self.assertEqual(self.p.textContent, 'a')
        self.assertIs(self.p.lastChild, self.c2)
        is_equal_nodes(self.p, [self.c1, 'a', self.c2])


class TestNonDocumentTypeChildNode(TestCase):
    def setUp(self):
        self.p = P()
        self.c1 = NDTC()
        self.c2 = NDTC()
        self.e1 = Element()
        self.e2 = Element()

    def test_element(self):
        self.p.appendChild(self.c1)
        self.assertIsNone(self.c1.previousElementSibling)
        self.assertIsNone(self.c1.nextElementSibling)
        self.p.appendChild(self.e1)
        self.assertIsNone(self.c1.previousElementSibling)
        self.assertIs(self.c1.nextElementSibling, self.e1)
        self.p.prepend(self.e2)
        self.assertIs(self.c1.previousElementSibling, self.e2)
        self.assertIs(self.c1.nextElementSibling, self.e1)

    def test_element2(self):
        self.p.appendChild(self.e1)
        self.assertIsNone(self.e1.previousElementSibling)
        self.assertIsNone(self.e1.nextElementSibling)


class TestChildNode(TestCase):
    def setUp(self):
        self.p = P()
        self.c1 = C()
        self.c2 = C()
        self.c3 = C()
        self.c4 = C()
        self.p.append(self.c1, self.c2)

    def test_after(self):
        self.c1.after(self.c3)
        is_equal_nodes(self.p, [self.c1, self.c3, self.c2])

        self.c1.after(self.c4)
        is_equal_nodes(self.p, [self.c1, self.c4, self.c3, self.c2])

    def test_after_text(self):
        self.c1.after('a')
        self.assertIs(self.p.firstChild, self.c1)
        self.assertEqual(self.p.textContent, 'a')
        is_equal_nodes(self.p, [self.c1, 'a', self.c2])

        self.c1.after('b', 'c')
        self.assertEqual(self.p.length, 5)
        self.assertEqual(self.p.textContent, 'bca')
        is_equal_nodes(self.p, [self.c1, 'b', 'c', 'a', self.c2])

    def test_after_multi(self):
        self.c1.after(self.c3, self.c4)
        is_equal_nodes(self.p, [self.c1, self.c3, self.c4, self.c2])

        self.c1.after(self.c2)
        # Is this behaviour correct?
        is_equal_nodes(self.p, [self.c1, self.c2, self.c3, self.c4])

    def test_arter_mixed_text_node(self):
        self.c1.after(self.c3, 'a', self.c4)
        is_equal_nodes(self.p, [self.c1, self.c3, 'a', self.c4, self.c2])

    def test_before(self):
        self.c1.before(self.c3)
        is_equal_nodes(self.p, [self.c3, self.c1, self.c2])

        self.c1.before(self.c4)
        is_equal_nodes(self.p, [self.c3, self.c4, self.c1, self.c2])

    def test_before_text(self):
        self.c1.before('a')
        self.assertEqual(self.p.textContent, 'a')
        is_equal_nodes(self.p, ['a', self.c1, self.c2])

        self.c1.before('b', 'c')
        is_equal_nodes(self.p, ['a', 'b', 'c', self.c1, self.c2])

    def test_before_multi(self):
        self.c2.before(self.c3, self.c4)
        is_equal_nodes(self.p, [self.c1, self.c3, self.c4, self.c2])

    def test_before_mixed(self):
        self.c2.before(self.c3, 'ab', self.c4, 'c')
        self.assertEqual(self.p.textContent, 'abc')
        is_equal_nodes(self.p, [self.c1, self.c3, 'ab', self.c4, 'c', self.c2])

    def test_replace(self):
        self.c1.replaceWith(self.c3)
        is_equal_nodes(self.p, [self.c3, self.c2])

    def test_replace_multi(self):
        self.c1.replaceWith(self.c3, self.c4)
        is_equal_nodes(self.p, [self.c3, self.c4, self.c2])

    def test_replace_text(self):
        self.c1.replaceWith('a', 'b')
        is_equal_nodes(self.p, ['a', 'b', self.c2])

    def test_remove(self):
        self.c2.remove()
        is_equal_nodes(self.p, [self.c1])


class TestCharacterData(TestCase):
    def setUp(self):
        self.node = Node()
        self.tnode = CharacterData('text')

    def test_textcontent(self):
        self.assertEqual(self.tnode.textContent, 'text')
        self.tnode.textContent = 'newtext'
        self.assertEqual(self.tnode.textContent, 'newtext')

    def test_length(self):
        self.assertEqual(self.tnode.length, 4)

    def test_append_data(self):
        self.tnode.appendData('new')
        self.assertEqual(self.tnode.textContent, 'textnew')

    def test_insert_data(self):
        self.tnode.insertData(1, 'new')
        self.assertEqual(self.tnode.textContent, 'tnewext')

    def test_delete_data(self):
        self.tnode.deleteData(1, 2)
        self.assertEqual(self.tnode.textContent, 'tt')

    def test_replace_data(self):
        self.tnode.replaceData(1, 2, 'new')
        self.assertEqual(self.tnode.textContent, 'tnewt')

    def test_invalid_methods(self):
        with self.assertRaises(NotImplementedError):
            self.tnode.appendChild(self.node)
        with self.assertRaises(NotImplementedError):
            self.tnode.removeChild(self.node)
        with self.assertRaises(NotImplementedError):
            self.tnode.insertBefore(self.node, self.node)
        with self.assertRaises(NotImplementedError):
            self.tnode.replaceChild(self.node, self.node)
        self.assertFalse(self.tnode.hasAttributes())
        self.assertFalse(self.tnode.hasChildNodes())
        self.assertEqual(len(self.tnode.childNodes), 0)


class TestText(TestCase):
    def setUp(self):
        self.node = Node()
        self.df = DocumentFragment()
        self.tnode = Text('text')

    def test_nodename(self):
        self.assertEqual(self.tnode.nodeName, '#text')

    def test_html_escape(self):
        self.assertEqual(self.tnode.html, 'text')
        self.tnode.textContent = '<'
        # not escape it has no parent
        self.assertEqual(self.tnode.html, '<')
        self.df.appendChild(self.tnode)
        # escape its parent is DocumentFragment or Element or...
        self.assertEqual(self.tnode.html, '&lt;')

        self.assertEqual(Text('<', parent=self.df).html, '&lt;')
        self.assertEqual(Text('>', parent=self.df).html, '&gt;')
        self.assertEqual(Text('&', parent=self.df).html, '&amp;')
        self.assertEqual(Text('"', parent=self.df).html, '&quot;')
        self.assertEqual(Text('\'', parent=self.df).html, '&#x27;')

    def test_appned(self):
        self.node.appendChild(self.tnode)
        self.assertTrue(self.node.hasChildNodes())
        self.assertIs(self.tnode.parentNode, self.node)

        node1 = Node()
        node2 = Node()
        self.node.appendChild(node1)
        self.assertIs(self.tnode.nextSibling, node1)
        self.assertIs(self.tnode, node1.previousSibling)

        self.node.insertBefore(node2, self.tnode)
        self.assertIs(self.tnode.previousSibling, node2)
        self.assertIs(self.tnode, node2.nextSibling)


class TestRawHtml(TestCase):
    def setUp(self):
        pass

    def test_rawhtml_content(self):
        rhtml = RawHtml('<a>')
        self.assertEqual(rhtml.html, '<a>')


class TestComment(TestCase):
    def setUp(self):
        self.c = Comment('comment')
        self.elm = P()

    def test_node_type(self):
        self.assertEqual(self.c.nodeType, self.c.COMMENT_NODE)
        self.assertEqual(self.c.nodeName, '#comment')

    def test_length(self):
        self.assertEqual(self.c.length, 7)

    def test_html(self):
        self.assertEqual('<!--comment-->', self.c.html)

    def test_append_comment(self):
        self.elm.appendChild(self.c)
        self.assertTrue(self.elm.hasChildNodes())
        self.assertEqual(self.elm.length, 1)
        self.assertEqual('<!--comment-->', self.elm.firstChild.html)


class TestDocumentType(TestCase):
    def setUp(self):
        self.dtype = DocumentType()
        self.node = Node()

    def test_nodename(self):
        self.assertEqual(self.dtype.nodeName, 'html')
        self.assertEqual(self.dtype.name, 'html')

    def test_parent(self):
        self.node.appendChild(self.dtype)
        self.assertIs(self.node, self.dtype.parentNode)

    def test_html(self):
        self.assertEqual(self.dtype.html, '<!DOCTYPE html>')


class TestDocumentFragment(TestCase):
    def setUp(self):
        self.df = DocumentFragment()
        self.elm = Element('a')
        self.c1 = Element('c1')
        self.c2 = Element('c2')
        self.c3 = Element('c3')

    def test_nodename(self):
        self.assertEqual(self.df.nodeName, '#document-fragment')

    def test_children(self):
        self.assertFalse(self.df.hasChildNodes())
        appended_child = self.df.appendChild(self.elm)
        self.assertEqual(self.df.html, '<a></a>')
        self.assertIs(appended_child, self.elm)

    def test_append_to_element(self):
        appended_child1 = self.df.appendChild(self.c1)
        appended_child2 = self.df.appendChild(self.c2)
        self.assertIs(self.df, self.c1.parentNode)
        self.assertIs(self.df, self.c2.parentNode)
        self.assertIs(appended_child1, self.c1)
        self.assertIs(appended_child2, self.c2)

        appended_df = self.elm.appendChild(self.df)
        self.assertIs(appended_df, self.df)
        self.assertEqual(self.df.length, 0)
        self.assertEqual(self.elm.length, 2)
        self.assertFalse(self.df.hasChildNodes())
        self.assertIsNone(self.df.parentNode)
        self.assertIs(self.elm, self.c1.parentNode)
        self.assertIs(self.elm, self.c2.parentNode)
        self.assertIsNone(self.df.parentNode)

    def test_insert_to_element(self):
        self.df.appendChild(self.c1)
        self.df.appendChild(self.c2)
        self.elm.appendChild(self.c3)

        inserted_node = self.elm.insertBefore(self.df, self.c3)
        self.assertIs(inserted_node, self.df)
        self.assertEqual(self.df.length, 0)
        self.assertEqual(self.elm.length, 3)
        self.assertFalse(self.df.hasChildNodes())
        self.assertIsNone(self.df.parentNode)
        self.assertIs(self.elm, self.c1.parentNode)
        self.assertIs(self.elm, self.c2.parentNode)

        self.assertIsNone(self.c1.previousSibling)
        self.assertIs(self.c1.nextSibling, self.c2)
        self.assertIs(self.c2.previousSibling, self.c1)
        self.assertIs(self.c2.nextSibling, self.c3)
        self.assertIs(self.c3.previousSibling, self.c2)
        self.assertIsNone(self.c3.nextSibling)
        self.assertIs(self.elm.firstChild, self.c1)
        self.assertIs(self.elm.lastChild, self.c3)

    def test_child(self):
        appended_child1 = self.df.appendChild(self.c1)
        appended_child2 = self.df.appendChild(self.c2)
        self.assertIs(appended_child1, self.c1)
        self.assertIs(appended_child2, self.c2)
        self.assertEqual(self.c1.html, '<c1></c1>')
        self.assertEqual(self.df.html, '<c1></c1><c2></c2>')

    def test_clone_node_sharrow(self):
        self.df.appendChild(self.c1)
        clone = self.df.cloneNode()
        self.assertEqual(self.df.length, 1)
        self.assertEqual(clone.length, 0)
        clone.appendChild(self.c2)
        self.assertEqual(self.df.length, 1)
        self.assertEqual(clone.length, 1)

    def test_clone_node_deep(self):
        self.df.appendChild(self.c1)
        clone = self.df.cloneNode(deep=True)
        self.assertEqual(self.df.length, 1)
        self.assertEqual(clone.length, 1)
        self.assertEqual(clone.html, '<c1></c1>')

        clone.appendChild(self.c2)
        self.assertEqual(self.df.length, 1)
        self.assertEqual(clone.length, 2)
        self.df.removeChild(self.c1)
        self.assertEqual(self.df.length, 0)
        self.assertEqual(clone.length, 2)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
Data binding example
'''

from wdom.themes.default import H1, Div, Input


class App(Div):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.text = H1(parent=self)
        self.text.textContent = 'Hello!'
        self.textbox = Input(parent=self)
        self.textbox.setAttribute('type', 'text')
        self.textbox.addEventListener('input', self.update)

    def update(self, event):
        self.text.textContent = self.textbox.getAttribute('value')


def sample_app(**kwargs):
    return App()


if __name__ == '__main__':
    import asyncio
    from wdom.document import get_document
    from wdom import server
    document = get_document()
    document.body.prepend(sample_app())
    server.start_server()
    try:
        asyncio.get_event_loop().run_forever()
    except KeyboardInterrupt:
        pass
    server.stop_server()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import unittest

from selenium.webdriver.common.keys import Keys

from wdom.tag import H1
from wdom.document import get_document
from wdom.misc import install_asyncio
from wdom.testing import WebDriverTestCase, TestCase, reset
from wdom.testing import close_webdriver


def setUpModule():
    install_asyncio()


def tearDownModule():
    close_webdriver()


class SimpleTestCase(WebDriverTestCase, TestCase):
    def setUp(self):
        super().setUp()
        reset()  # I don't know why, but need reset for aiohttp.
        document = get_document()
        self.h1 = H1()
        self.h1.textContent = 'TITLE'
        document.body.appendChild(self.h1)
        self.start()

    def test_page(self):
        tag = self.wd.find_element_by_css_selector(
            '[rimo_id="{}"]'.format(self.h1.rimo_id))
        self.assertEqual(tag.text, 'TITLE')


class TestDataBinding(WebDriverTestCase, TestCase):
    def setUp(self):
        super().setUp()
        reset()  # I don't know why, but need reset for aiohttp.
        from wdom.examples.data_binding import sample_app
        document = get_document()
        document.body.prepend(sample_app())
        self.start()

    @unittest.skipIf(os.environ.get('TRAVIS', False),
                     reason='This test not pass only on travis')
    def test_app(self):
        view = self.wd.find_element_by_tag_name('h1')
        self.assertEqual(view.text, 'Hello!')
        input = self.wd.find_element_by_tag_name('input')
        self.send_keys(input, 'abcde')
        self.wait_until(lambda: view.text == 'abcde')
        self.assertEqual(view.text, 'abcde')
        for i in range(5):
            self.send_keys(input, Keys.BACKSPACE)
        self.wait_until(lambda: view.text == '')
        self.assertEqual(view.text, '')
        self.send_keys(input, 'new')
        self.wait_until(lambda: view.text == 'new')
        self.assertEqual(view.text, 'new')


class TestRevText(WebDriverTestCase, TestCase):
    def setUp(self):
        super().setUp()
        reset()  # I don't know why, but need reset for aiohttp.
        from wdom.examples.rev_text import sample_app
        document = get_document()
        document.body.prepend(sample_app())
        self.start()

    @unittest.skipIf(os.environ.get('TRAVIS', False),
                     reason='This test not pass only on travis')
    def test_app(self):
        view = self.wd.find_element_by_tag_name('h1')
        text = 'Click!'
        self.assertEqual(view.text, text)
        # need long wait time
        self.wait(times=20)
        view.click()
        self.wait(times=20)
        self.wait_until(lambda: view.text == text[::-1])
        self.assertEqual(view.text, text[::-1])
        self.wait(times=20)
        view.click()
        self.wait(times=20)
        self.wait_until(lambda: view.text == text)
        self.assertEqual(view.text, text)


class TestTimer(WebDriverTestCase, TestCase):
    def setUp(self):
        super().setUp()
        reset()  # I don't know why, but need reset for aiohttp.
        from wdom.examples.timer import sample_app
        document = get_document()
        document.body.prepend(sample_app())
        self.start()

    @unittest.skipIf(os.environ.get('TRAVIS', False),
                     reason='This test not pass only on travis')
    def test_timer(self):
        view = self.wd.find_element_by_tag_name('h1')
        start_btn = self.wd.find_element_by_id('start_btn')
        stop_btn = self.wd.find_element_by_id('stop_btn')
        reset_btn = self.wd.find_element_by_id('reset_btn')
        self.wait(times=20)

        def test_timer():
            self.assertEqual(view.text, '180.00')
            start_btn.click()
            self.wait(times=20)
            self.assertTrue(float(view.text) < 179.90)
            stop_btn.click()
            t = view.text
            self.wait(times=20)
            self.assertEqual(view.text, t)

        test_timer()
        reset_btn.click()
        self.wait(times=20)
        test_timer()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import re
import asyncio
import unittest
from unittest.mock import MagicMock

from syncer import sync

from wdom.testing import TestCase
from wdom.document import get_document
from wdom.misc import install_asyncio
from wdom.node import DocumentFragment, Text
from wdom.web_node import WebElement
from wdom.testing import RemoteBrowserTestCase, NoSuchElementException
from wdom.testing import start_remote_browser, close_remote_browser


def setUpModule():
    install_asyncio()
    start_remote_browser()


def tearDownModule():
    close_remote_browser()


class ElementTestCase(RemoteBrowserTestCase):
    if os.environ.get('TRAVIS', False):
        wait_time = 0.1

    def setUp(self):
        super().setUp()
        self.document = get_document()
        self.document.body.prepend(self.get_elements())
        self.start()

    def get_elements(self):
        raise NotImplementedError


class TestWebElement(ElementTestCase, TestCase):
    def get_elements(self):
        self.root = WebElement('div')
        self.tag = WebElement('span', parent=self.root)
        self.df = DocumentFragment()
        self.c1 = WebElement('c1')
        self.c2 = WebElement('c2')
        self.c3 = WebElement('c3')
        self.c4 = WebElement('c4')
        self.c1.textContent = 'child1'
        self.c2.textContent = 'child2'
        self.c3.textContent = 'child3'
        self.c4.textContent = 'child4'
        return self.root

    def test_connection(self):
        self.wait_until(lambda: self.root.connected)
        self.assertTrue(self.root.connected)
        self.browser.get('http://example.com/')
        self.assertIsFalse(self.root.connected)

    def test_text_content(self):
        self.set_element(self.tag)
        self.assertEqual(self.element.text, '')
        self.tag.textContent = 'text'
        self.wait_until(lambda: self.element.text == 'text')
        self.assertEqual(self.element.text, 'text')

        self.c1.textContent = 'child'
        self.tag.appendChild(self.c1)
        self.wait_until(lambda: self.element.text == 'textchild')
        self.assertEqual(self.element.text, 'textchild')

        self.tag.textContent = 'NewText'
        self.wait_until(lambda: self.element.text == 'NewText')
        self.assertEqual(self.element.text, 'NewText')
        with self.assertRaises(NoSuchElementException):
            self.set_element(self.c1, self.wait_time * 10)

        self.set_element(self.tag)
        t_node = Text('TextNode')
        self.tag.replaceChild(t_node, self.tag.childNodes[0])
        self.wait_until(lambda: self.element.text == 'TextNode')
        self.assertEqual(self.element.text, 'TextNode')

        self.tag.removeChild(self.tag.childNodes[0])
        self.set_element(self.tag)
        self.wait_until(lambda: self.element.text == '')
        self.assertEqual(self.element.text, '')

    def test_attr(self):
        self.set_element(self.tag)
        self.assertIsNone(self.element.get_attribute('src'))
        self.tag.setAttribute('src', 'a')
        self.wait_until(lambda: self.element.get_attribute('src') == 'a')
        self.assertEqual(self.element.get_attribute('src'), 'a')
        self.tag.removeAttribute('src')
        self.wait_until(lambda: self.element.get_attribute('src') is None)
        self.assertIsNone(self.element.get_attribute('src'))

    def test_addremove_child(self):
        self.tag.appendChild(self.c1)
        self.set_element(self.c1)
        self.assertEqual(self.element.text, 'child1')
        self.c1.textContent = 'Child'
        self.wait_until(lambda: self.element.text == 'Child')
        self.assertEqual(self.element.text, 'Child')

        self.set_element(self.tag)
        self.wait_until(lambda: self.element.text == 'Child')
        self.assertEqual(self.element.text, 'Child')

        self.tag.removeChild(self.c1)
        with self.assertRaises(NoSuchElementException):
            self.wait(0.1)
            self.set_element(self.c1, self.wait_time * 10)

        self.set_element(self.tag)
        self.wait_until(lambda: self.element.text == '')
        self.assertEqual(self.element.text, '')

    def test_insert_child(self):
        self.set_element(self.tag)
        # test parent in constructor
        self.c1 = WebElement('c1', parent=self.tag)
        self.c1.textContent = 'child1'

        self.assertIsTrue(self.set_element(self.c1))
        with self.assertRaises(NoSuchElementException):
            self.wait(0.1)
            self.set_element(self.c2, self.wait_time * 10)

        self.set_element(self.tag)
        self.wait_until(lambda: self.element.text == 'child1')
        self.assertEqual(self.element.text, 'child1')

        self.tag.insertBefore(self.c2, self.c1)
        self.assertIsTrue(self.set_element(self.c2))

        self.set_element(self.tag)
        self.wait_until(lambda: self.element.text == 'child2child1')
        self.assertEqual(self.element.text, 'child2child1')

        self.tag.empty()
        self.wait_until(lambda: self.element.text == '')
        self.assertEqual(self.element.text, '')
        with self.assertRaises(NoSuchElementException):
            self.set_element(self.c1, self.wait_time * 10)
        with self.assertRaises(NoSuchElementException):
            self.set_element(self.c2, self.wait_time * 10)

    def test_add_df(self):
        self.set_element(self.tag)
        self.df.append(self.c1, self.c2, 'text')
        self.tag.appendChild(self.df)
        self.wait_until(lambda: self.element.text == 'child1child2text')
        self.assertEqual(self.element.text, 'child1child2text')

        df = DocumentFragment()
        df.append(self.c3, 'text2')
        self.tag.appendChild(df)
        self.wait_until(
            lambda: self.element.text == 'child1child2textchild3text2')
        self.assertEqual(self.element.text, 'child1child2textchild3text2')

    def test_insert_df(self):
        self.set_element(self.tag)
        self.tag.appendChild(self.c1)
        self.df.append(self.c2, self.c3, 'text')
        self.tag.insertBefore(self.df, self.c1)
        self.wait_until(lambda: self.element.text == 'child2child3textchild1')
        self.assertEqual(self.element.text, 'child2child3textchild1')

        df = DocumentFragment()
        df.append(self.c4, 'text2')
        self.tag.insertBefore(df, self.c3)
        self.wait_until(
            lambda: self.element.text == 'child2child4text2child3textchild1')
        self.assertEqual(self.element.text,
                         'child2child4text2child3textchild1')

    def test_replace_child(self):
        self.set_element(self.tag)
        self.tag.appendChild(self.c1)

        with self.assertRaises(NoSuchElementException):
            self.set_element(self.c2, self.wait_time * 10)
        self.assertIsTrue(self.set_element(self.c1))
        self.wait_until(lambda: self.element.text == 'child1')
        self.assertEqual(self.element.text, 'child1')

        self.tag.replaceChild(self.c2, self.c1)
        with self.assertRaises(NoSuchElementException):
            self.wait(0.1)
            self.set_element(self.c1, self.wait_time * 10)
        self.assertIsTrue(self.set_element(self.c2))
        self.wait_until(lambda: self.element.text == 'child2')
        self.assertEqual(self.element.text, 'child2')
        self.set_element(self.tag)
        self.wait_until(lambda: self.element.text == 'child2')
        self.assertEqual(self.element.text, 'child2')

    def test_append(self):
        self.set_element(self.tag)
        self.tag.append(self.c1)
        self.wait_until(lambda: self.element.text == 'child1')
        self.assertEqual(self.element.text, 'child1')

        self.tag.append(self.c2, self.c3)
        self.wait_until(lambda: self.element.text == 'child1child2child3')
        self.assertEqual(self.element.text, 'child1child2child3')

        self.tag.append(self.c4, self.c1)
        self.wait_until(
            lambda: self.element.text == 'child2child3child4child1')
        self.assertEqual(self.element.text, 'child2child3child4child1')

        self.tag.append('t1', 't2')
        self.wait_until(
            lambda: self.element.text == 'child2child3child4child1t1t2')
        self.assertEqual(self.element.text, 'child2child3child4child1t1t2')

    def test_prepend(self):
        self.set_element(self.tag)
        self.tag.prepend(self.c1)
        self.wait_until(lambda: self.element.text == 'child1')
        self.assertEqual(self.element.text, 'child1')

        self.tag.prepend(self.c2, self.c3)
        self.wait_until(lambda: self.element.text == 'child2child3child1')
        self.assertEqual(self.element.text, 'child2child3child1')

        self.tag.prepend(self.c4, self.c1)
        self.wait_until(
            lambda: self.element.text == 'child4child1child2child3')
        self.assertEqual(self.element.text, 'child4child1child2child3')

        self.tag.prepend('t1', 't2')
        self.wait_until(
            lambda: self.element.text == 't1t2child4child1child2child3')
        self.assertEqual(self.element.text, 't1t2child4child1child2child3')

    def test_prepend_append_text(self):
        self.set_element(self.tag)
        self.tag.append('t1')
        self.wait_until(lambda: self.element.text == 't1')
        self.assertEqual(self.element.text, 't1')

        self.tag.firstChild.remove()
        self.wait_until(lambda: self.element.text == '')
        self.assertEqual(self.element.text, '')

        self.tag.prepend('t2')
        self.wait_until(lambda: self.element.text == 't2')
        self.assertEqual(self.element.text, 't2')

        self.tag.append('t3', 't4')
        self.wait_until(lambda: self.element.text == 't2t3t4')
        self.assertEqual(self.element.text, 't2t3t4')

        self.tag.prepend('t5', 't6')
        self.wait_until(lambda: self.element.text == 't5t6t2t3t4')
        self.assertEqual(self.element.text, 't5t6t2t3t4')

    def test_after(self):
        self.set_element(self.tag)
        self.tag.append(self.c1)
        self.c1.after(self.c2)
        self.wait_until(lambda: self.element.text == 'child1child2')
        self.assertEqual(self.element.text, 'child1child2')

        self.c1.after(self.c3, self.c4)
        self.wait_until(
            lambda: self.element.text == 'child1child3child4child2')
        self.assertEqual(self.element.text, 'child1child3child4child2')

        self.c1.after(self.c2, 'text')
        self.wait_until(
            lambda: self.element.text == 'child1child2textchild3child4')
        self.assertEqual(self.element.text, 'child1child2textchild3child4')

    def test_before(self):
        self.set_element(self.tag)
        self.tag.append(self.c1)
        self.c1.before(self.c2)
        self.wait_until(lambda: self.element.text == 'child2child1')
        self.assertEqual(self.element.text, 'child2child1')

        self.c1.before(self.c3, self.c4)
        self.wait_until(
            lambda: self.element.text == 'child2child3child4child1')
        self.assertEqual(self.element.text, 'child2child3child4child1')

        self.c1.before(self.c2, 'text')
        self.wait_until(
            lambda: self.element.text == 'child3child4child2textchild1')
        self.assertEqual(self.element.text, 'child3child4child2textchild1')

    def test_after_before_text(self):
        self.set_element(self.tag)
        self.tag.append('a')
        t = self.tag.firstChild
        t.after('b')
        self.wait_until(lambda: self.element.text == 'ab')
        self.assertEqual(self.element.text, 'ab')

        t.after('c', 'd')
        self.wait_until(lambda: self.element.text == 'acdb')
        self.assertEqual(self.element.text, 'acdb')

        t.before('e')
        self.wait_until(lambda: self.element.text == 'eacdb')
        self.assertEqual(self.element.text, 'eacdb')

        t.before('f', 'g')
        self.wait_until(lambda: self.element.text == 'efgacdb')
        self.assertEqual(self.element.text, 'efgacdb')

    def test_shortcut_attr(self):
        self.tag.textContent = 'TAG'
        self.set_element(self.tag)
        self.wait_until(lambda: self.element.is_displayed())
        self.assertIsTrue(self.element.is_displayed())
        self.tag.hidden = True
        self.wait_until(lambda: not self.element.is_displayed())
        self.assertIsFalse(self.element.is_displayed())
        self.tag.hidden = False
        self.wait_until(lambda: self.element.is_displayed())
        self.assertIsTrue(self.element.is_displayed())

    def test_style(self):
        self.tag.textContent = 'Style'
        self.set_element(self.tag)
        self.wait_until(lambda: self.element.get_attribute('style') == '')
        self.assertEqual(self.element.get_attribute('style'), '')
        style = 'color: red;'
        self.tag.style = style
        self.wait_until(lambda: self.element.get_attribute('style') == style)
        self.assertEqual(self.element.get_attribute('style'), style)
        self.tag.style.color = 'black'
        self.wait_until(
            lambda: self.element.get_attribute('style') == 'color: black;')
        self.assertEqual(self.element.get_attribute('style'), 'color: black;')

    def test_classlist(self):
        self.set_element(self.tag)
        self.assertEqual(self.element.get_attribute('class'), '')
        self.tag.classList.add('a')
        self.wait_until(lambda: self.element.get_attribute('class') == 'a')
        self.assertEqual(self.element.get_attribute('class'), 'a')
        self.tag.classList.add('b', 'c', 'd')
        self.wait_until(
            lambda: self.element.get_attribute('class') == 'a b c d')
        self.assertEqual(self.element.get_attribute('class'), 'a b c d')

        self.tag.classList.remove('c')
        self.wait_until(lambda: self.element.get_attribute('class') == 'a b d')
        self.assertEqual(self.element.get_attribute('class'), 'a b d')
        self.tag.classList.remove('a', 'd')
        self.wait_until(lambda: self.element.get_attribute('class') == 'b')
        self.assertEqual(self.element.get_attribute('class'), 'b')

        self.tag.classList.toggle('b')
        self.wait_until(lambda: self.element.get_attribute('class') == '')
        self.assertEqual(self.element.get_attribute('class'), '')
        self.tag.classList.toggle('b')
        self.wait_until(lambda: self.element.get_attribute('class') == 'b')
        self.assertEqual(self.element.get_attribute('class'), 'b')

    def test_click(self):
        mock = MagicMock(_is_coroutine=False)
        self.tag.addEventListener('click', mock)
        self.tag.click()
        self.wait_until(lambda: mock.call_count == 1)
        self.assertEqual(mock.call_count, 1)

    @sync
    @asyncio.coroutine
    def test_get_rect(self):
        rect = WebElement('div', style='width:200px;height:100px;')
        self.tag.appendChild(rect)
        yield from asyncio.sleep(self.wait_time)

        data = yield from rect.getBoundingClientRect()
        self.assertEqual(data['width'], 200)
        self.assertEqual(data['height'], 100)

    @sync
    @asyncio.coroutine
    def test_scroll(self):
        rect = WebElement('div',
                          style='width:3000px;height:3000px;background:#eee;')
        self.tag.appendChild(rect)
        yield from asyncio.sleep(self.wait_time)

        X = yield from rect.scrollX()
        Y = yield from rect.scrollY()
        self.assertEqual(X['x'], 0)
        self.assertEqual(Y['y'], 0)

        rect.scrollTo(200, 200)
        yield from asyncio.sleep(self.wait_time)
        X = yield from rect.scrollX()
        Y = yield from rect.scrollY()
        self.assertEqual(X['x'], 200)
        self.assertEqual(Y['y'], 200)

    def test_exec(self):
        self.tag.exec('this.style = "color: red;"')
        self.set_element(self.tag)
        self.wait_until(
            lambda: re.search(r'255,\s*0,\s* 0,\s*1\s*',
                self.element.value_of_css_property('color')))
        self.assertRegex(self.element.value_of_css_property('color'),
                         r'255,\s*0,\s* 0,\s*1\s*')

        self.tag.exec('node.style = "color: blue;"')
        self.wait_until(
            lambda: re.search(r'0,\s*0,\s* 255,\s*1\s*',
                self.element.value_of_css_property('color')))
        self.assertRegex(self.element.value_of_css_property('color'),
                         r'0,\s*0,\s*255,\s*1\s*')

    def test_exec_error(self):
        with self.assertLogs('wdom.server', 'ERROR') as log:
            self.tag.exec('a.b')
            self.wait_until(lambda: len(log.output) > 0)
        self.assertRegex(log.output[0], r'JS: ReferenceError')


class TestEvent(ElementTestCase, TestCase):
    def get_elements(self):
        self.root = WebElement('div')
        self.tag = WebElement('span', parent=self.root)

        self.click_event_mock = MagicMock()
        self.click_event_mock._is_coroutine = False

        self.btn = WebElement('button')
        self.btn.textContent = 'click'
        self.btn.addEventListener('click', self.click_event_mock)

        self.input_event_mock = MagicMock()
        self.input_event_mock._is_coroutine = False

        self.input = WebElement('input', type='text')
        self.input.addEventListener('input', self.input_event_mock)

        self.root.appendChild(self.btn)
        self.root.appendChild(self.input)
        return self.root

    @unittest.skipIf(os.environ.get('TRAVIS', False),
                     reason='This test not pass only on travis')
    def test_click(self):
        # need to wait some times
        self.wait(times=10)
        self.set_element(self.btn)
        self.wait(times=10)
        self.element.click()
        self.wait(times=10)
        self.wait_until(lambda: self.click_event_mock.call_count == 1)
        self.assertEqual(self.click_event_mock.call_count, 1)

    def test_input(self):
        self.set_element(self.input)
        self.element.send_keys('abc')
        self.wait_until(lambda: self.input_event_mock.call_count == 3)
        self.assertEqual(self.input_event_mock.call_count, 3)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys
import time
import logging
import asyncio
import unittest
from multiprocessing import Process, Pipe
from types import FunctionType, MethodType

from selenium import webdriver
from selenium.common.exceptions import NoSuchElementException
from selenium.webdriver.common.utils import free_port
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.support.select import Select

from tornado.httpclient import AsyncHTTPClient, HTTPResponse
from tornado.platform.asyncio import AsyncIOMainLoop, to_asyncio_future
from tornado.websocket import websocket_connect, WebSocketClientConnection

from wdom.misc import install_asyncio
from wdom import options
from wdom.window import customElements
from wdom.element import Element
from wdom import server

driver = webdriver.Firefox
local_webdriver = None
remote_webdriver = None
browser_implict_wait = 0
logger = logging.getLogger(__name__)


def reset():
    """Reset all wdom objects.

    This function clear all connections, elements, and resistered custom
    elements. This function also makes new document/application and set them.
    """
    from wdom.document import get_new_document, set_document
    set_document(get_new_document())
    from wdom.server import _tornado
    _tornado.connections.clear()
    _tornado.set_application(_tornado.Application())
    try:
        from wdom.server import _aiohttp
        _aiohttp.connections.clear()
        _aiohttp.set_application(_aiohttp.Application())
    except ImportError:
        pass
    Element._elements_with_id.clear()
    Element._elements.clear()
    customElements.clear()


def suppress_logging():
    """Suppress log output to stdout.

    This function is intended to be used in test's setup. This function removes
    log handler of ``wdom`` logger and set NullHandler to suppress log.
    """
    options.root_logger.removeHandler(options._log_handler)
    options.root_logger.addHandler(logging.NullHandler())


class TestCase(unittest.TestCase):
    """Base class for testing wdom modules.

    This class is a sub class of the ``unittest.TestCase``. After all test
    methods, reset wdom's global objects like document, application, and
    elements. If you use ``tearDown`` method, do not forget to call
    ``super().tearDown()``.

    If you want to reuse document/application object in your test class, please
    set them in each setup phase as follow::

        @classmethod
        def setUpClass(cls):
            cls.your_doc = get_document()
            cls.your_app = get_app()

        def setUp(self):
            from wdom.document import set_document
            from wdom.server import set_application
            set_document(self.your_doc)
            set_application(self.your_app)
    """

    def tearDown(self):
        reset()
        super().tearDown()

    def assertIsTrue(self, bl):
        """Check arg is exactly True, not truthy."""
        self.assertIs(bl, True)

    def assertIsFalse(self, bl):
        """Check arg is exactly False, not falsy."""
        self.assertIs(bl, False)


class HTTPTestCase(TestCase):
    """For http/ws connection test."""

    wait_time = 0.01
    timeout = 1.0
    _server_started = False
    _ws_connections = []

    def start(self):
        """Start web server.

        Please call this method after prepraring document.
        """
        with self.assertLogs('wdom', 'INFO'):
            self.server = server.start_server(port=0)
        self.port = self.server.port
        self.url = 'http://localhost:{}'.format(self.port)
        self.ws_url = 'ws://localhost:{}'.format(self.port)
        self._server_started = True

    def tearDown(self):
        """Terminate server and close all ws client connections."""
        if self._server_started:
            with self.assertLogs('wdom', 'INFO'):
                server.stop_server(self.server)
            self._server_started = False
        while self._ws_connections:
            ws = self._ws_connections.pop()
            ws.close()
        super().tearDown()

    @asyncio.coroutine
    def fetch(self, url: str, encoding: str = 'utf-8') -> HTTPResponse:
        """Fetch url and return ``tornado.httpclient.HTTPResponse`` object.

        Response body is decoded by ``encoding`` and set ``text`` property of
        the response. If failed to decode, ``text`` property will be ``None``.
        """
        response = yield from to_asyncio_future(
            AsyncHTTPClient().fetch(url, raise_error=False))
        try:
            response.text = response.body.decode(encoding)
        except UnicodeDecodeError:
            response.text = None
        return response

    @asyncio.coroutine
    def ws_connect(self, url: str, timeout: float = None
                   ) -> WebSocketClientConnection:
        """Make WebSocket connection to the url.

        Retries up to _max (default: 20) times. Client connections made by this
        method are closed after each test method.
        """
        st = time.perf_counter()
        timeout = timeout or self.timeout
        while (time.perf_counter() - st) < timeout:
            try:
                ws = yield from to_asyncio_future(websocket_connect(url))
            except ConnectionRefusedError:
                yield from self.wait()
                continue
            else:
                self._ws_connections.append(ws)
                return ws
        raise ConnectionRefusedError(
            'WebSocket connection refused: {}'.format(url))

    @asyncio.coroutine
    def wait(self, timeout: float = None, times: int = 1):
        """Coroutine to wait for ``timeout``.

        ``timeout`` is second to wait, and its default value is
        ``self.wait_time``. If ``times`` are specified, wait for
        ``timeout * times``.
        """
        for i in range(times):
            yield from asyncio.sleep(timeout or self.wait_time)


def start_webdriver():
    """Start WebDriver and set implicit_wait if it is not started."""
    global local_webdriver
    if local_webdriver is None:
        local_webdriver = driver()
        if browser_implict_wait:
            local_webdriver.implicitly_wait(browser_implict_wait)


def close_webdriver():
    """Close WebDriver."""
    global local_webdriver
    if local_webdriver is not None:
        local_webdriver.close()
        local_webdriver = None


def get_webdriver():
    """Return WebDriver of current process.

    If it is not started yet, start and return it.
    """
    if globals().get('local_webdriver') is None:
        start_webdriver()
    return local_webdriver


def _clear():
    global conn, wd_conn, browser_manager, remote_webdriver
    conn = None
    wd_conn = None
    browser_manager = None
    remote_webdriver = None


def start_remote_browser():
    """Start remote browser process."""
    global browser_manager, conn, wd_conn
    conn, wd_conn = Pipe()

    def start_browser():
        global wd_conn
        bc = BrowserController(wd_conn)
        bc.run()

    browser_manager = Process(target=start_browser)
    browser_manager.start()


def close_remote_browser():
    """Terminate remote browser process."""
    global conn, browser_manager
    conn.send({'target': 'process', 'method': 'quit'})
    time.sleep(0.3)
    logger.info('\nRemote Browser closed')
    conn.close()
    if browser_manager is not None:
        browser_manager.terminate()
    _clear()


def get_remote_browser():
    """Start new WebDriver for remote process."""
    global remote_webdriver
    if remote_webdriver is None:
        remote_webdriver = driver()
        if browser_implict_wait:
            remote_webdriver.implicitly_wait(browser_implict_wait)
        return remote_webdriver
    else:
        return remote_webdriver


class BrowserController:
    """Class to run and wrap webdriver in different proceess.
    """
    _select_methods = [s for s in dir(Select) if not s.startswith('_')]

    def __init__(self, conn):
        """Set up connection and start webdriver.

        ``conn`` is a one end of ``Pipe()``, which is used the inter-process
        communication.
        """
        self.conn = conn
        self.wd = get_remote_browser()
        self.element = None

    def set_element_by_id(self, id):
        """Find element with ``id`` and set it to element property.

        When successfully find the element, send ``True``. If failed to find the
        element, send message ``Error NoSuchElement: {{ id }}``.
        """
        try:
            self.element = self.wd.find_element_by_css_selector(
                '[rimo_id="{}"]'.format(id))
            return True
        except NoSuchElementException:
            return 'Error NoSuchElement: ' + id

    def quit(self, *args):
        """Terminate WebDriver."""
        self.wd.quit()
        return 'closed'

    def close(self, *args):
        """Close WebDriver."""
        self.wd.close()
        return 'closed'

    def _execute_method(self, method, args):
        if isinstance(method, (FunctionType, MethodType)):
            self.conn.send(method(*args))
        else:
            # not callable, just send it back
            self.conn.send(method)

    def run(self):
        """Wait message from the other end of the connection.

        When gat message, execute the method specified by the message. The
        message should be a python's dict, which must have ``target`` and
        ``method`` field.
        """
        while True:
            req = self.conn.recv()
            target = req.get('target', '')
            method_name = req.get('method', '')
            args = req.get('args', [])
            if target == 'process':
                method = getattr(self, method_name)
            elif target == 'browser':
                method = getattr(self.wd, method_name)
            elif target == 'element':
                if self.element is None:
                    # Element must be set
                    self.conn.send('Error: No Element Set')
                    continue
                if (method_name in self._select_methods and
                        self.element.tag_name.lower() == 'select'):
                    s = Select(self.element)
                    method = getattr(s, method_name)
                else:
                    method = getattr(self.element, method_name)
            self._execute_method(method, args)


def wait_for():
    """Wait the response from the remote process and return it."""
    return asyncio.get_event_loop().run_until_complete(wait_coro())


@asyncio.coroutine
def wait_coro():
    """Wait response from the other process."""
    while True:
        state = conn.poll()
        if state:
            res = conn.recv()
            return res
        else:
            yield from asyncio.sleep(0)
            continue


def _get_properties(cls):
    props = set()
    for k, v in vars(cls).items():
        if not isinstance(v, (FunctionType, MethodType)):
            props.add(k)
    return props


class Controller:
    """Base class for remote browser controller."""
    target = None
    properties = set()

    def __getattr__(self, attr: str):
        """Call methods related to this controller."""
        global conn

        def wrapper(*args):
            conn.send({'target': self.target, 'method': attr, 'args': args})
            res = wait_for()
            if isinstance(res, str):
                if res.startswith('Error NoSuchElement'):
                    raise NoSuchElementException(res)
                elif res.startswith('Error'):
                    raise ValueError(res)
            return res
        if attr in self.properties:
            return wrapper()
        else:
            return wrapper


class ProcessController(Controller):
    """Controller of remote browser process."""
    target = 'process'


class RemoteBrowserController(Controller):
    """Controller of remote web driver."""
    target = 'browser'
    properties = _get_properties(WebDriver)


class RemoteElementController(Controller):
    """Controller of remote web driver element."""
    target = 'element'
    properties = _get_properties(WebElement)


class TimeoutError(Exception):
    """The operation is not completed by timeout."""


class RemoteBrowserTestCase:
    """This class is **Experimental**.

    Utility class for testing apps with webdriver in another process. Mainly
    used for development and test of wdom library itself. This class does not
    support all methods provided by selenium.webdriver, but maybe enough.

    After seting up your document, call ``start`` method in setup sequence.
    """
    #: seconds to wait for by ``wait`` method.
    wait_time = 0.01
    #: secondes for deault timeout for ``wait_until`` method
    timeout = 1.0

    def start(self):
        """Start remote browser process."""
        self._prev_logging = options.config.logging
        options.config.logging = 'warn'
        self.proc = ProcessController()
        self.browser = RemoteBrowserController()
        self.element = RemoteElementController()
        try:
            self.server = server.start_server(port=0)
        except OSError:
            self.wait(0.2)
            self.server = server.start_server(port=0)
        self.address = self.server.address
        self.url = 'http://{0}:{1}/'.format(self.address, self.port)
        self.browser.get(self.url)
        self.wait_until(lambda: server.is_connected())

    def tearDown(self):
        options.config.logging = self._prev_logging
        server.stop_server()
        sys.stdout.flush()
        sys.stderr.flush()
        super().tearDown()

    @property
    def port(self) -> int:
        """Get port of the server."""
        return self.server.port

    def wait(self, timeout: float = None, times: int = 1):
        """Wait for ``timeout`` seconds.

        Default timeout is ``RemoteBrowserTestCase.wait_time``.
        """
        loop = asyncio.get_event_loop()
        for i in range(times):
            loop.run_until_complete(asyncio.sleep(timeout or self.wait_time))

    def wait_until(self, func, timeout=None):
        """Wait until ``func`` returns True or exceeds timeout.

        ``func`` is called with no argument. Unit of ``timeout`` is second, and
        its default value is RemoteBrowserTestCase.timeout class variable
        (default: 1.0).
        """
        st = time.perf_counter()
        timeout = timeout or self.timeout
        while (time.perf_counter() - st) < timeout:
            if func():
                return
            self.wait()
        raise TimeoutError('{} did not return True until timeout'.format(func))

    def _set_element(self, node):
        try:
            res = self.proc.set_element_by_id(node.rimo_id)
            return res
        except NoSuchElementException:
            return False

    def set_element(self, node, timeout=None):
        """Set the ``node`` as a target node of the remote browser process."""
        try:
            self.wait_until(lambda: self._set_element(node), timeout)
            return True
        except TimeoutError:
            pass
        raise NoSuchElementException('element not found: {}'.format(node))


class WebDriverTestCase:
    """Base class for testing UI on browser.

    This class starts up an HTTP server on a new subprocess.

    Subclasses should call ``start`` method after seting up your document.
    After ``start`` method called, the web server is running on the other
    process so you cannot make change on the document. If you need to change
    document after server started, please use ``RemoteBrowserTestCase`` class
    instead.
    """
    #: seconds to wait for by ``wait`` method.
    wait_time = 0.01
    #: secondes for deault timeout for ``wait_until`` method
    timeout = 1.0

    @classmethod
    def setUpClass(cls):
        cls._orig_loop = asyncio.get_event_loop()
        # Need to use different loop for aiohttp after remote_browser tests
        # but I can't understand why...?
        asyncio.set_event_loop(asyncio.new_event_loop())
        # When change default loop, tornado's ioloop needs to be reinstalled
        AsyncIOMainLoop().clear_current()
        AsyncIOMainLoop().clear_instance()
        install_asyncio()
        reset()

    @classmethod
    def tearDownClass(cls):
        # Set original loop and reinstall to tornado
        asyncio.set_event_loop(cls._orig_loop)
        AsyncIOMainLoop().clear_current()
        AsyncIOMainLoop().clear_instance()
        install_asyncio()
        reset()

    def start(self):
        """Start server and web driver."""
        self.wd = get_webdriver()

        def start_server(port):
            import asyncio
            from wdom import server
            server.start_server(port=port)
            asyncio.get_event_loop().run_forever()

        self.address = 'localhost'
        self.port = free_port()
        self.url = 'http://{0}:{1}/'.format(self.address, self.port)

        self.server = Process(
            target=start_server,
            args=(self.port, )
        )
        self.server.start()
        self.wait(times=10)
        self.wd.get(self.url)

    def tearDown(self):
        """Terminate server subprocess."""
        self.server.terminate()
        sys.stdout.flush()
        sys.stderr.flush()
        self.wait(times=10)
        super().tearDown()

    def wait(self, timeout: float = None, times: int = 1):
        """Wait for ``timeout`` or ``self.wait_time``."""
        loop = asyncio.get_event_loop()
        for i in range(times):
            loop.run_until_complete(asyncio.sleep(timeout or self.wait_time))

    def wait_until(self, func, timeout=None):
        """Wait until ``func`` returns True or exceeds timeout.

        ``func`` is called with no argument. Unit of ``timeout`` is second, and
        its default value is RemoteBrowserTestCase.timeout class variable
        (default: 1.0).
        """
        st = time.perf_counter()
        timeout = timeout or self.timeout
        while (time.perf_counter() - st) < timeout:
            if func():
                return
            self.wait()
        raise TimeoutError('{} did not return True until timeout'.format(func))

    def send_keys(self, element, keys: str):
        """Send ``keys`` to ``element`` one-by-one.

        Safer than using ``element.send_keys`` method.
        """
        for k in keys:
            element.send_keys(k)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from wdom.themes.default import H1, Div


class App(Div):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.text = H1('Click!', parent=self)
        self.text.addEventListener('click', self.reverse)

    def reverse(self, event):
        self.text.textContent = self.text.textContent[::-1]


def sample_app(**kwargs) -> Div:
    return App()


if __name__ == '__main__':
    import asyncio
    from wdom.document import get_document
    from wdom import server
    document = get_document()
    document.body.prepend(sample_app())
    server.start_server()
    try:
        asyncio.get_event_loop().run_forever()
    except KeyboardInterrupt:
        pass
    server.stop_server()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import misaka as m
from pygments import highlight
from pygments.styles import get_all_styles
from pygments.formatters import HtmlFormatter
from pygments.lexers import get_lexer_by_name

from wdom.document import get_document
from wdom.themes.bootstrap3 import css_files, js_files
from wdom.themes.bootstrap3 import Div, TextArea, Col6, Row, H1, Hr
from wdom.themes.bootstrap3 import Select, Option, Style


src = '''
## Source Code Example

```py
from collections import OrderedDict

class MyDict(OrderedDict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        print('Create my dict')
```
'''


class HighlighterRenderer(m.HtmlRenderer):
    def blockcode(self, text, lang):
        if not lang:
            return '\n<pre><code>{}</code></pre>\n'.format(text.strip())
        lexer = get_lexer_by_name(lang)
        formatter = HtmlFormatter()
        return highlight(text, lexer, formatter)


class Editor(Row):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.md = m.Markdown(HighlighterRenderer(), extensions=('fenced-code',))
        self.css = Style(parent=self)

        self.setAttribute('style', 'height: 80vh;')
        editor_col = Col6(parent=self)
        self.editor = TextArea(parent=editor_col)
        self.editor.setAttribute('style', 'height: 80vh')

        viewer_col = Col6(parent=self)

        style_selector = Select(parent=viewer_col)
        styles = sorted(get_all_styles())
        styles.remove('default')
        style_selector.appendChild(
            Option('default', value='default', selected=True))
        for style in styles:
            style_selector.appendChild(Option(style, value=style))
        style_selector.addEventListener(
            'change', lambda event: self.set_style(
                event.currentTarget['value']))

        self.viewer = Div(parent=viewer_col)
        self.viewer.setAttribute(
            'style',
            '''
            height: 100%;
            min-height: 80vh;
            padding: 1em 2em;
            border: 1px solid #ddd;
            border-radius: 3px;
            ''',
        )

        self.editor.addEventListener('input', self.render)
        self.editor.addEventListener('change', self.render)

        self.set_style('default')
        self.editor.textContent = src
        self.viewer.innerHTML = self.md(src)

    def render(self, event):
        self.viewer.innerHTML = self.md(event.currentTarget['value'])

    def set_style(self, style:str):
        self.css.innerHTML = HtmlFormatter(style=style).get_style_defs()


def sample_page(**kwargs):
    doc = get_document(**kwargs)
    for js in js_files:
        doc.add_jsfile(js)
    for css in css_files:
        doc.add_cssfile(css)
    app = Div(parent=doc.body, style='width: 90vw; margin: auto')
    title = H1('Simple Markdown Editor', class_='text-center')
    app.appendChild(title)
    app.appendChild(Hr())
    app.appendChild(Editor())
    return doc



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import re
import logging

__all__ = ('logger', 'theme_list')

logger = logging.getLogger(__name__)
curdir = os.path.dirname(os.path.abspath(__file__))
theme_list = sorted(set(re.sub(r'\.py[codx]?$', '', file)
                        for file in os.listdir(curdir)
                        if not file.startswith('_')))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import json
import logging
from asyncio import Future
from typing import Optional
from xml.dom import Node

from wdom import server
from wdom.interface import Event

logger = logging.getLogger(__name__)


class WebIF:
    @property
    def ownerDocument(self) -> Optional[Node]:
        return None

    def __init__(self, *args, **kwargs):
        self._reqid = 0
        self._tasks = {}
        super().__init__(*args, **kwargs)

    @property
    def connected(self) -> bool:
        '''When this instance has any connection, return True.'''
        return server.is_connected()

    def on_event_pre(self, event: Event):
        '''Hook executed before dispatching events.
        Used for set values changed by user input, in some elements like input,
        textarea, or select.
        In this method, event.currentTarget is a dict sent from browser.
        '''
        pass

    def on_response(self, msg):
        response = msg.get('data', False)
        if response:
            task = self._tasks.pop(msg.get('reqid'), False)
            if task and not task.cancelled() and not task.done():
                task.set_result(msg.get('data'))

    def js_exec(self, method: str, *args):
        '''Execute ``method`` in the related node on browser, via web socket
        connection. Other keyword arguments are passed to ``params`` attribute.
        If this node is not in any document tree (namely, this node does not
        have parent node), the ``method`` is not executed.
        '''
        if self.connected:
            self.ws_send(dict(method=method, params=args))

    def js_query(self, query) -> Future:
        if self.connected:
            self.js_exec(query, self._reqid)
            fut = Future()
            self._tasks[self._reqid] = fut
            self._reqid += 1
            return fut
        else:
            fut = Future()
            fut.set_result(None)
            return fut

    def ws_send(self, obj):
        '''Send message to the related nodes on browser, with ``tagname`` and
        ``id`` which specifies relation between python's object and element
        on browser. The message is serialized by JSON object and send via
        WebSocket connection.
        '''
        obj['target'] = 'node'
        obj['id'] = self.rimo_id
        obj['tag'] = self.tag
        msg = json.dumps(obj)
        server.send_message(msg)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from wdom.tag import NewTagClass as NewTag
from wdom.tag import *

name = 'Blaze'
project_url = 'http://blazecss.com/'
project_repository = 'https://github.com/BlazeCSS/blaze'
license = 'MIT License'
license_url = 'https://github.com/BlazeCSS/blaze/blob/master/LICENSE'

css_files = [
    '//cdn.jsdelivr.net/blazecss/2.1.1/blaze.min.css',
]

Button = NewTag('Button', bases=Button, class_='c-button')
DefaultButton = NewTag('DefaultButton', 'button', Button, is_='default-button')
PrimaryButton = NewTag('PrimaryButton', 'button', Button, class_='c-button--primary', is_='primary-button')
SecondaryButton = NewTag('SecondaryButton', 'button', Button, class_='c-button--secondary', is_='secondary-button')
SuccessButton = NewTag('SuccessButton', 'button', Button, class_='c-button--success', is_='success-button')
InfoButton = NewTag('InfoButton', 'button', Button, class_='c-button--primary', is_='info-button')
WarningButton = NewTag('WarningButton', 'button', Button, class_='c-button--secondary', is_='warining-button')
DangerButton = NewTag('DangerButton', 'button', Button, class_='c-button--error', is_='danger-button')
ErrorButton = NewTag('ErrorButton', 'button', Button, class_='c-button--error', is_='error-button')
LinkButton = NewTag('LinkButton', 'button', Button, is_='c-link-button')

Input = NewTag('Input', 'input', Input, class_='c-field')
TextInput = NewTag('TextInput', 'input', TextInput, class_='c-field')
Textarea = NewTag('Textarea', 'textarea', Textarea, class_='c-field')
Select = NewTag('Select', 'select', Select, class_='c-choice')

Ul = NewTag('Ul', 'ul', Ul, class_='c-list')
Ol = NewTag('Ol', 'ol', Ol, class_='c-list--ordered')
Li = NewTag('Li', 'li', Li, class_='c-list__item')

Table = NewTag('Table', 'table', Table, class_='c-table')
Tr = NewTag('Tr', 'tr', Tr, class_='c-table__row')
Th = NewTag('Th', 'th', Th, class_='c-table__cell')
Td = NewTag('Td', 'td', Td, class_='c-table__cell')

H1 = NewTag('H1', 'div', H1, class_='c-heading c-heading--super')
H2 = NewTag('H2', 'div', H2, class_='c-heading c-heading--xlarge')
H3 = NewTag('H3', 'div', H3, class_='c-heading c-heading--large')
H4 = NewTag('H4', 'div', H4, class_='c-heading c-heading--medium')
H5 = NewTag('H5', 'div', H5, class_='c-heading c-heading--small')
H6 = NewTag('H6', 'div', H6, class_='c-heading c-heading--xsmall')

Row = NewTag('Row', 'div', Row, class_='o-gird')
Col = NewTag('Col', 'div', Col, class_='o-grid__cell')
Col1 = NewTag('Col1', 'div', Col, class_='o-grid__cell--width-10', is_='col1')
Col2 = NewTag('Col2', 'div', Col, class_='o-grid__cell--width-15', is_='col2')
Col3 = NewTag('Col3', 'div', Col, class_='o-grid__cell--width-25', is_='col3')
Col4 = NewTag('Col4', 'div', Col, class_='o-grid__cell--width-33', is_='col4')
Col5 = NewTag('Col5', 'div', Col, class_='o-grid__cell--width-40', is_='col5')
Col6 = NewTag('Col6', 'div', Col, class_='o-grid__cell--width-50', is_='col6')
Col7 = NewTag('Col7', 'div', Col, class_='o-grid__cell--width-60', is_='col7')
Col8 = NewTag('Col8', 'div', Col, class_='o-grid__cell--width-66', is_='col8')
Col9 = NewTag('Col9', 'div', Col, class_='o-grid__cell--width-75', is_='col9')
Col10 = NewTag('Col10', 'div', Col, class_='o-grid__cell--width-85', is_='col10')
Col11 = NewTag('Col11', 'div', Col, class_='o-grid__cell--width-90', is_='col11')
Col12 = NewTag('Col12', 'div', Col, class_='o-grid__cell--width-100', is_='col12')

extended_classes = [
    Button,
    DefaultButton,
    PrimaryButton,
    SecondaryButton,
    SuccessButton,
    InfoButton,
    WarningButton,
    DangerButton,
    ErrorButton,
    LinkButton,
    Input,
    TextInput,
    Textarea,
    Select,
    Ul,
    Ol,
    Li,
    Table,
    Tr,
    Th,
    Td,
    H1,
    H2,
    H3,
    H4,
    H5,
    H6,
    Row,
    Col,
    Col1,
    Col2,
    Col3,
    Col4,
    Col5,
    Col6,
    Col7,
    Col8,
    Col9,
    Col10,
    Col11,
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import tempfile
import shutil
from functools import partial
from types import ModuleType
from typing import Optional, Union, Callable
import weakref

from wdom.options import config
from wdom.interface import Event
from wdom.node import Node, DocumentType, Text, RawHtml, Comment
from wdom.node import DocumentFragment
from wdom.element import Element, Attr, _create_element
from wdom.web_node import WebElement
from wdom.tag import HTMLElement
from wdom.tag import Html, Head, Body, Meta, Link, Title, Script
from wdom.window import Window


def getElementById(id: Union[str, int]) -> Optional[Node]:
    elm = Element._elements_with_id.get(str(id))
    if elm and elm.ownerDocument:
        return elm
    else:
        return None


def getElementByRimoId(id: Union[str, int]) -> Optional[WebElement]:
    elm = WebElement._elements_with_rimo_id.get(str(id))
    if elm and elm.ownerDocument:
        return elm
    else:
        return None


def _cleanup(path):
    if os.path.isdir(path):
        shutil.rmtree(path)


class Document(Node):
    nodeType = Node.DOCUMENT_NODE
    nodeName = '#document'

    @property
    def defaultView(self) -> Window:
        return self._window

    @property
    def tempdir(self) -> str:
        return self._tempdir

    def __init__(self, doctype='html', title='W-DOM', charset='utf-8',
                 default_class=HTMLElement, autoreload=None, reload_wait=None):
        self._tempdir = _tempdir = tempfile.mkdtemp()
        self._finalizer = weakref.finalize(self, partial(_cleanup, _tempdir))
        super().__init__()
        self._window = Window(self)
        self._default_class = default_class
        self._autoreload = autoreload
        self._reload_wait = reload_wait

        self.doctype = DocumentType(doctype, parent=self)
        self.html = Html(parent=self)
        self.head = Head(parent=self.html)
        self.charset_element = Meta(parent=self.head)
        self.characterSet = charset
        self.title_element = Title(parent=self.head)
        self.title = title

        self.body = Body(parent=self.html)
        self.script = Script(parent=self.body)
        self._autoreload_script = Script(parent=self.head)

    def _set_autoreload(self):
        self._autoreload_script.textContent = ''
        if self._autoreload is None:
            autoreload = (config.autoreload or config.debug)
        else:
            autoreload = self._autoreload

        if autoreload:
            ar_script = []
            ar_script.append('var RIMO_AUTORELOAD = true')
            if self._reload_wait is not None:
                ar_script.append('var RIMO_RELOAD_WAIT = {}'.format(
                    self._reload_wait))
            self._autoreload_script.textContent = '\n{}\n'.format(
                '\n'.join(ar_script))

    def getElementById(self, id: Union[str, int]) -> Optional[Node]:
        elm = getElementById(id)
        if elm and elm.ownerDocument is self:
            return elm

    def getElementByRimoId(self, id: Union[str, int]) -> Optional[WebElement]:
        elm = getElementByRimoId(id)
        if elm and elm.ownerDocument is self:
            return elm

    def createElement(self, tag: str):
        return _create_element(tag, base=self._default_class)

    def createDocumentFragment(self):
        return DocumentFragment()

    def createTextNode(self, text: str):
        return Text(text)

    def createComment(self, text: str):
        return Comment(text)

    def createEvent(self, event: str):
        return Event(event)

    def createAttribute(self, name: str):
        return Attr(name)

    @property
    def title(self) -> str:
        return self.title_element.textContent

    @title.setter
    def title(self, value: str):
        self.title_element.textContent = value

    @property
    def characterSet(self) -> str:
        return self.charset_element.getAttribute('charset')

    @characterSet.setter
    def characterSet(self, value: str):
        self.charset_element.setAttribute('charset', value)

    @property
    def charset(self) -> str:
        return self.characterSet

    @charset.setter
    def charset(self, value: str):
        self.characterSet = value

    def add_jsfile(self, src: str):
        self.body.appendChild(Script(src=src))

    def add_jsfile_head(self, src: str):
        self.head.appendChild(Script(src=src))

    def add_cssfile(self, src: str):
        self.head.appendChild(Link(rel='stylesheet', href=src))

    def add_header(self, header: str):
        self.head.appendChild(RawHtml(header))

    def register_theme(self, theme: ModuleType) -> None:
        if not hasattr(theme, 'css_files'):
            raise ValueError('theme module must include `css_files`.')
        for css in getattr(theme, 'css_files', []):
            self.add_cssfile(css)
        for js in getattr(theme, 'js_files', []):
            self.add_jsfile(js)
        for header in getattr(theme, 'headers', []):
            self.add_header(header)
        for cls in getattr(theme, 'extended_classes', []):
            self.defaultView.customElements.define(cls)

    def build(self) -> str:
        self._set_autoreload()
        return ''.join(child.html for child in self.childNodes)


def get_new_document(include_rimo: bool = True,
                     include_skeleton: bool = False,
                     include_normalizecss: bool = False,
                     autoreload: Optional[bool] = None,
                     reload_wait: int = None,
                     log_level: int = None,
                     log_prefix: str = None,
                     log_console: bool = False,
                     ws_url: str = None,
                     document_factory: Callable[..., Document] = Document,
                     **kwargs) -> Document:
    document = document_factory(
        autoreload=autoreload,
        reload_wait=reload_wait,
        **kwargs)
    if log_level is None:
        log_level = config.logging

    log_script = []
    if log_level is not None:
        if isinstance(log_level, str):
            log_script.append('var RIMO_LOG_LEVEL = \'{}\''.format(log_level))
        elif isinstance(log_level, int):
            log_script.append('var RIMO_LOG_LEVEL = {}'.format(log_level))
    if log_prefix is not None:
        log_script.append('var RIMO_LOG_PREFIX = \'{}\''.format(log_prefix))
    if log_console:
        log_script.append('var RIMO_LOG_CONSOLE = true')
    if log_script:
        _s = Script(parent=document.head)
        _s.textContent = '\n{}\n'.format('\n'.join(log_script))

    if ws_url is not None:
        _s = Script(parent=document.head)
        _s.textContent = '\nvar RIMO_WS_URL = \'{}\'\n'.format(ws_url)

    if include_rimo:
        document.add_jsfile_head('_static/js/rimo/rimo.js')

    return document


# get_document = get_new_document
def get_document(*args, **kwargs):
    return rootDocument


def set_document(new_document: Document, *args, **kwargs):
    global rootDocument
    rootDocument = new_document


rootDocument = get_new_document()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import importlib

from nose_parameterized import parameterized

from wdom.themes import theme_list
from wdom.testing import TestCase


class TestThemesImport(TestCase):
    @parameterized.expand(theme_list)
    def test_import_all(self, theme):
        a = importlib.import_module('wdom.themes.' + theme)
        self.assertIn('name', dir(a))
        self.assertIn('project_url', dir(a))
        self.assertIn('project_repository', dir(a))
        self.assertIn('license', dir(a))
        self.assertIn('license_url', dir(a))
        self.assertIn('css_files', dir(a))
        self.assertIn('js_files', dir(a))
        self.assertIn('headers', dir(a))
        self.assertIn('extended_classes', dir(a))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from wdom.tag import NewTagClass as NewTag
from wdom.tag import *

name = 'MUI'
project_url = 'https://www.muicss.com/'
project_repository = 'https://github.com/muicss/mui'
license = 'MIT License'
license_url = 'https://github.com/muicss/mui/blob/master/LICENSE.txt'

css_files = [
    '//cdn.muicss.com/mui-0.5.1/css/mui.min.css',
]
js_files = [
    '//cdn.muicss.com/mui-0.5.1/js/mui.min.js',
]
headers = []

Button = NewTag('Button', bases=Button, class_='mui-btn')
RaisedButton = NewTag('RaisedButton', 'button', Button, class_='mui-btn--raised')
FlatButton = NewTag('FlatButton', 'button', Button, class_='mui-btn--flat')
DefaultButton = NewTag('DefaultButton', 'button', RaisedButton, is_='default-button')
PrimaryButton = NewTag('PrimaryButton', 'button', RaisedButton, class_='mui-btn--primary', is_='primary-button')
SecondaryButton = NewTag('SecondaryButton', 'button', RaisedButton, class_='mui-btn--accent', is_='secondary-button')
SuccessButton = NewTag('SuccessButton', 'button', RaisedButton, class_='mui-btn--accent', is_='success-button')
InfoButton = NewTag('InfoButton', 'button', FlatButton, class_='mui-btn--accent', is_='info-button')
WarningButton = NewTag('WarningButton', 'button', Button, class_='mui-btn--danger', is_='warning-button')
DangerButton = NewTag('DangerButton', 'button', WarningButton, class_='mui-btn--raised', is_='danger-button')
ErrorButton = NewTag('ErrorButton', 'button', WarningButton, class_='mui-btn--raised', is_='error-button')
LinkButton = NewTag('LinkButton', 'button', FlatButton, class_='mui-btn--primary', is_='link-button')

Table = NewTag('Table', 'table', Table, class_='mui-table')

H1 = NewTag('H1', 'div', H1, class_='mui--text-display4')
H2 = NewTag('H2', 'div', H2, class_='mui--text-display3')
H3 = NewTag('H3', 'div', H3, class_='mui--text-display2')
H4 = NewTag('H4', 'div', H4, class_='mui--text-display1')
H5 = NewTag('H5', 'div', H5, class_='mui--text-headline')
H6 = NewTag('H6', 'div', H6, class_='mui--text-title')

Container = NewTag('Container', 'div', Container, class_='mui-container')
Wrapper = NewTag('Wrapper', 'div', Wrapper, class_='mui-container')
Row = NewTag('Row', 'div', Row, class_='mui-row')

Col1 = NewTag('Col1', 'div', Col1, class_='mui-col-xs-1 mui-col-sm-1 mui-col-md-1 mui-col-lg-1')
Col2 = NewTag('Col2', 'div', Col2, class_='mui-col-xs-2 mui-col-sm-2 mui-col-md-2 mui-col-lg-2')
Col3 = NewTag('Col3', 'div', Col3, class_='mui-col-xs-3 mui-col-sm-3 mui-col-md-3 mui-col-lg-3')
Col4 = NewTag('Col4', 'div', Col4, class_='mui-col-xs-4 mui-col-sm-4 mui-col-md-4 mui-col-lg-4')
Col5 = NewTag('Col5', 'div', Col5, class_='mui-col-xs-5 mui-col-sm-5 mui-col-md-5 mui-col-lg-5')
Col6 = NewTag('Col6', 'div', Col6, class_='mui-col-xs-6 mui-col-sm-6 mui-col-md-6 mui-col-lg-6')
Col7 = NewTag('Col7', 'div', Col7, class_='mui-col-xs-7 mui-col-sm-7 mui-col-md-7 mui-col-lg-7')
Col8 = NewTag('Col8', 'div', Col8, class_='mui-col-xs-8 mui-col-sm-8 mui-col-md-8 mui-col-lg-8')
Col9 = NewTag('Col9', 'div', Col9, class_='mui-col-xs-9 mui-col-sm-9 mui-col-md-9 mui-col-lg-9')
Col10 = NewTag('Col10', 'div', Col10, class_='mui-col-xs-10 mui-col-sm-10 mui-col-md-10 mui-col-lg-10')
Col11 = NewTag('Col11', 'div', Col11, class_='mui-col-xs-11 mui-col-sm-11 mui-col-md-11 mui-col-lg-11')

extended_classes = [
    Button,
    RaisedButton,
    FlatButton,
    DefaultButton,
    PrimaryButton,
    SecondaryButton,
    SuccessButton,
    InfoButton,
    WarningButton,
    DangerButton,
    ErrorButton,
    LinkButton,
    Table,
    H1,
    H2,
    H3,
    H4,
    H5,
    H6,
    Container,
    Wrapper,
    Row,
    Col1,
    Col2,
    Col3,
    Col4,
    Col5,
    Col6,
    Col7,
    Col8,
    Col9,
    Col10,
    Col11,
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from wdom.themes.bootstrap3 import Tag, H1, Button, Div
from wdom.themes.bootstrap3 import Container, FormGroup, TextInput
from wdom.themes.bootstrap3 import css_files, js_files
from wdom.document import get_document, Document


class Item(Tag):
    tag = 'span'
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)


def sample_page(**kwargs) -> Document:
    app = Container()
    title = H1(parent=app)
    title.textContent = 'Todo example'
    form = FormGroup(parent=app)
    text_input = TextInput()
    form.append(text_input)
    add_button = Button(parent=form)
    add_button.textContent = 'ADD'
    todo_list = Div(parent=app)
    todo_heading = Div(parent=todo_list)
    todo_heading.append('Todo list')
    # add_button.addEventListener('click')

    def new_item(event=None) -> Tag:
        item = Item()
        item.append('New Item')
        todo_list.append(item)
        return item

    add_button.addEventListener('click', new_item)

    page = get_document(app=app, **kwargs)
    for css in css_files:
        page.add_cssfile(css)
    for js in js_files:
        page.add_jsfile(js)

    return page



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys
import gc
from unittest import skipIf

from wdom.css import CSSStyleDeclaration
from wdom.node import Text
from wdom.element import (
    DOMTokenList, NamedNodeMap, Attr, Element, HTMLElement,
    HTMLSelectElement, HTMLOptionElement,
)
from wdom.window import customElements
from wdom.testing import TestCase


class TestDOMTokenList(TestCase):
    def setUp(self):
        self.tokens = DOMTokenList(self)

    def test_add(self):
        self.assertEqual(self.tokens.length, 0)
        self.assertEqual(self.tokens.toString(), '')

        self.tokens.add('a')
        self.assertEqual(self.tokens.length, 1)
        self.assertEqual(self.tokens.toString(), 'a')

        self.tokens.add('a')
        self.assertEqual(self.tokens.length, 1)
        self.assertEqual(self.tokens.toString(), 'a')

        self.tokens.add('b')
        self.assertEqual(self.tokens.length, 2)
        self.assertEqual(self.tokens.toString(), 'a b')

        with self.assertRaises(ValueError):
            self.tokens.add('a c')
        with self.assertRaises(ValueError):
            self.tokens.add('a', 'b c')

    def test_remove(self):
        self.tokens.add('a')
        self.tokens.add('b')
        self.assertEqual(self.tokens.length, 2)

        self.tokens.remove('a')
        self.assertEqual(self.tokens.length, 1)
        self.assertEqual(self.tokens.toString(), 'b')

        self.tokens.remove('a')
        self.assertEqual(self.tokens.length, 1)
        self.assertEqual(self.tokens.toString(), 'b')

        self.tokens.remove('b')
        self.assertEqual(self.tokens.length, 0)
        self.assertEqual(self.tokens.toString(), '')

        with self.assertRaises(ValueError):
            self.tokens.remove('a c')

    def test_toggle(self):
        self.tokens.toggle('a')
        self.assertEqual(self.tokens.length, 1)
        self.assertEqual(self.tokens.toString(), 'a')

        self.tokens.toggle('b')
        self.assertEqual(self.tokens.length, 2)
        self.assertEqual(self.tokens.toString(), 'a b')

        self.tokens.toggle('a')
        self.assertEqual(self.tokens.length, 1)
        self.assertEqual(self.tokens.toString(), 'b')

        with self.assertRaises(ValueError):
            self.tokens.toggle('a c')

    def test_item(self):
        self.assertIsNone(self.tokens.item(1))
        self.tokens.add('a')
        self.assertIsNone(self.tokens.item(1))
        self.tokens.add('b')
        self.assertEqual(self.tokens.item(1), 'b')

    def test_contains(self):
        self.assertFalse(self.tokens.contains('a'))
        self.tokens.add('a')
        self.assertTrue(self.tokens.contains('a'))

        with self.assertRaises(ValueError):
            self.tokens.contains('a c')

    def test_add_multi(self):
        self.tokens.add('a', 'b')
        self.assertEqual(len(self.tokens), 2)
        self.assertEqual('a b', self.tokens.toString())
        self.tokens.remove('a')
        self.assertEqual(len(self.tokens), 1)
        self.assertEqual('b', self.tokens.toString())

    def test_add_multi_string(self):
        # used at initialization of Element
        self.tokens._append('a b')
        self.assertEqual(len(self.tokens), 2)
        self.assertEqual('a b', self.tokens.toString())
        self.tokens.remove('a')
        self.assertEqual(len(self.tokens), 1)
        self.assertEqual('b', self.tokens.toString())

    def test_add_multi_list(self):
        # used at initialization of Element
        self.tokens._append(['a', 'b'])
        self.assertEqual(len(self.tokens), 2)
        self.assertEqual('a b', self.tokens.toString())
        self.tokens.remove('a')
        self.assertEqual(len(self.tokens), 1)
        self.assertEqual('b', self.tokens.toString())

    def test_add_multi_mixed(self):
        # used at initialization of Element
        self.tokens._append(['a', 'b c'])
        self.assertEqual(len(self.tokens), 3)
        self.assertEqual('a b c', self.tokens.toString())
        self.tokens.remove('b')
        self.assertEqual(len(self.tokens), 2)
        self.assertEqual('a c', self.tokens.toString())

    def test_remove_multi(self):
        self.tokens.add('a', 'b', 'c')
        self.tokens.remove('a', 'c')
        self.assertEqual(self.tokens.length, 1)
        self.assertNotIn('a', self.tokens)
        self.assertIn('b', self.tokens)
        self.assertNotIn('c', self.tokens)

    def test_add_none(self):
        with self.assertRaises(TypeError):
            self.tokens.add(None)

    def test_add_blank(self):
        self.tokens.add('')
        self.assertEqual(len(self.tokens), 0)
        self.assertFalse(bool(self.tokens))
        self.assertEqual('', self.tokens.toString())

    def test_add_invlalid(self):
        with self.assertRaises(TypeError):
            self.tokens.add(1)
        with self.assertRaises(TypeError):
            self.tokens.add(Element('a'))
        self.assertEqual(len(self.tokens), 0)
        self.assertFalse(bool(self.tokens))
        self.assertEqual('', self.tokens.toString())

    def test_iter(self):
        cls = ['a', 'b', 'c']
        self.tokens.add(*cls)
        for c in self.tokens:
            self.assertIn(c, cls)
            cls.remove(c)
        self.assertEqual(len(cls), 0)


class TestAttr(TestCase):
    def setUp(self):
        self.id = Attr('id')
        self.cls = Attr('class')
        self.src = Attr('src')

    def test_name(self):
        self.assertEqual(self.id.name, 'id')
        self.assertEqual(self.cls.name, 'class')
        self.assertEqual(self.src.name, 'src')

    def test_value(self):
        self.src.value = 'a'
        self.assertEqual(self.src.value, 'a')

    def test_html(self):
        self.src.value = 'a'
        self.assertEqual(self.src.html, 'src="a"')

    def test_escape(self):
        self.src.value = '"a"'
        self.assertEqual(self.src.html, 'src="&quot;a&quot;"')

    @skipIf(True, 'Now Attr node need owner element to check boolean')
    def test_boolean_attr(self):
        hidden = Attr('hidden')
        hidden.value = True
        self.assertEqual(hidden.html, 'hidden')
        hidden.value = False
        self.assertEqual(hidden.html, '')

    def test_isid(self):
        self.assertTrue(self.id.isId)
        self.assertFalse(self.cls.isId)
        self.assertFalse(self.src.isId)


class TestNamedNodeMap(TestCase):
    def setUp(self):
        self.map = NamedNodeMap(self)
        self.attr = Attr('src', value='a')

    def test_addremove(self):
        self.assertEqual(self.map.length, 0)
        self.map.setNamedItem(self.attr)
        self.assertEqual(self.map.length, 1)
        self.assertEqual(self.map.getNamedItem('src').value, 'a')
        self.assertIsNone(self.map.getNamedItem('aaa'))
        self.map.removeNamedItem(Attr('aaa'))
        self.assertEqual(self.map.length, 1)
        self.map.removeNamedItem(Attr('src'))
        self.assertEqual(self.map.length, 0)

    def test_item(self):
        self.map.setNamedItem(self.attr)
        self.assertIsNone(self.map.item(1))
        self.assertIs(self.map.item(0), self.attr)


class TestElementMeta(TestCase):
    def setUp(self):
        class NewTag(Element):
            _special_attr_string = ['a', 'b']
            _special_attr_boolean = ['c', 'd']
        self.tag = NewTag('a')

    def check_noattr(self, attr):
        self.assertNotIn(attr, self.tag.attributes)
        self.assertFalse(self.tag.hasAttribute(attr))
        self.assertIsNone(self.tag.getAttribute(attr))

    def check_hasattr(self, attr):
        self.assertIn(attr, self.tag.attributes)
        self.assertTrue(self.tag.hasAttribute(attr))

    def test_special_attr_str(self):
        for attr in ['a', 'b', 'c', 'd', 'e']:
            self.check_noattr(attr)
        self.assertEqual(self.tag.a, '')
        self.assertEqual(self.tag.b, '')
        self.assertEqual(self.tag.html, '<a></a>')

    def test_special_attr_str_setter(self):
        self.tag.a = 'a'
        self.check_hasattr('a')
        self.assertEqual(self.tag.a, 'a')
        self.assertEqual(self.tag.getAttribute('a'), 'a')
        self.assertEqual(self.tag.html, '<a a="a"></a>')

        self.tag.removeAttribute('a')
        self.check_noattr('a')
        self.assertEqual(self.tag.a, '')
        self.assertEqual(self.tag.html, '<a></a>')

    def test_special_attr_str_setattr(self):
        self.tag.setAttribute('b', 'b')
        self.check_hasattr('b')
        self.assertEqual(self.tag.b, 'b')
        self.assertEqual(self.tag.getAttribute('b'), 'b')
        self.assertEqual(self.tag.html, '<a b="b"></a>')

        del self.tag.b
        self.check_noattr('b')
        self.assertEqual(self.tag.b, '')
        self.assertEqual(self.tag.html, '<a></a>')
        # delete again, but not raise error
        del self.tag.b

    def test_special_attr_bool(self):
        for attr in ['a', 'b', 'c', 'd', 'e']:
            self.check_noattr(attr)
        self.assertFalse(self.tag.a)
        self.assertFalse(self.tag.b)
        self.assertEqual(self.tag.html, '<a></a>')

    def test_special_attr_bool_setter(self):
        self.tag.c = True
        self.check_hasattr('c')
        self.assertTrue(self.tag.c)
        self.assertEqual(self.tag.html, '<a c></a>')

        self.tag.c = False
        self.check_noattr('c')
        self.assertFalse(self.tag.c)
        self.assertEqual(self.tag.html, '<a></a>')

    def test_special_attr_bool_setattr(self):
        self.tag.setAttribute('d', 'd')
        self.check_hasattr('d')
        self.assertTrue(self.tag.d)
        self.assertEqual(self.tag.html, '<a d></a>')

        self.tag.removeAttribute('d')
        self.check_noattr('d')
        self.assertFalse(self.tag.d)
        self.assertEqual(self.tag.html, '<a></a>')

    def test_special_attr_bool_setattr_empty(self):
        # In this case, 'd' still exists and True (compatible with JS...)
        self.tag.setAttribute('d', '')
        self.check_hasattr('d')
        self.assertEqual(self.tag.html, '<a d></a>')

        # Now remove 'd'
        self.tag.d = ''
        self.check_noattr('d')
        self.assertFalse(self.tag.d)
        self.assertEqual(self.tag.html, '<a></a>')

    def test_special_attr_bool_setattr_false(self):
        # In this case, 'd' still exists and True (compatible with JS...)
        self.tag.setAttribute('d', False)
        self.check_hasattr('d')
        self.assertEqual(self.tag.html, '<a d></a>')

        # Now remove 'd'
        del self.tag.d
        self.check_noattr('d')
        self.assertFalse(self.tag.d)
        self.assertEqual(self.tag.html, '<a></a>')


class TestElement(TestCase):
    def setUp(self):
        customElements.clear()
        self.elm = Element('tag')
        self.c1 = Element('c1')
        self.c2 = Element('c2')
        self.c1.classList.add('c1')
        self.c2.classList.add('c2')

    def tearDown(self):
        del self.elm
        del self.c1
        del self.c2
        super().tearDown()

    def test_constructor(self):
        elm = Element('a')
        self.assertEqual(elm.nodeName, 'A')
        self.assertEqual(elm.tagName, 'A')
        self.assertEqual(elm.localName, 'a')

        self.assertFalse(elm.hasChildNodes())
        self.assertFalse(elm.hasAttributes())

    def test_init_parent(self):
        elm = Element('a', parent=self.elm)
        self.assertIs(elm.parentNode, self.elm)

    def test_init_attrs(self):
        elm = Element('a', src='b', href='c')
        self.assertFalse(elm.hasChildNodes())
        self.assertTrue(elm.hasAttributes())
        self.assertTrue(elm.getAttribute('src'), 'b')
        self.assertTrue(elm.getAttribute('href'), 'c')

    def test_attrs(self):
        self.assertFalse(self.elm.hasAttributes())
        self.assertFalse(self.elm.hasAttribute('src'))
        self.assertNotIn('src', self.elm.attributes)
        self.assertIsNone(self.elm.getAttribute('src'))
        self.assertEqual(self.elm.html, '<tag></tag>')
        self.elm.setAttribute('src', 'a')
        self.assertTrue(self.elm.hasAttributes())
        self.assertTrue(self.elm.hasAttribute('src'))
        self.assertIn('src', self.elm.attributes)
        self.assertEqual(self.elm.getAttribute('src'), 'a')
        self.assertEqual(self.elm.html, '<tag src="a"></tag>')

        self.elm.removeAttribute('src')
        self.assertFalse(self.elm.hasAttributes())
        self.assertFalse(self.elm.hasAttribute('src'))
        self.assertNotIn('src', self.elm.attributes)
        self.assertIsNone(self.elm.getAttribute('src'))
        self.assertEqual(self.elm.html, '<tag></tag>')

    def test_id(self):
        self.assertEqual(self.elm.id, '')
        self.elm.setAttribute('id', 'a')
        self.assertEqual(self.elm.getAttribute('id'), 'a')
        self.assertEqual(self.elm.id, 'a')
        self.elm.id = 'b'
        self.assertEqual(self.elm.getAttribute('id'), 'b')
        self.assertEqual(self.elm.id, 'b')

    def test_class_list(self):
        self.assertIsNone(self.elm.getAttribute('class'))
        self.assertFalse(self.elm.hasAttribute('class'))
        self.assertFalse(self.elm.hasAttributes())

        self.elm.setAttribute('class', 'a')
        self.assertEqual(self.elm.getAttribute('class'), 'a')
        self.assertTrue(self.elm.hasAttribute('class'))
        self.assertTrue(self.elm.hasAttributes())

        self.elm.removeAttribute('class')
        self.assertIsNone(self.elm.getAttribute('class'))
        self.assertFalse(self.elm.hasAttribute('class'))
        self.assertFalse(self.elm.hasAttributes())

    def test_start_tag(self):
        self.assertEqual(self.elm.start_tag, '<tag>')
        self.elm.setAttribute('src', 'a')
        self.assertEqual(self.elm.start_tag, '<tag src="a">')

        self.elm.setAttribute('class', 'b')
        self.assertEqual(self.elm.start_tag, '<tag src="a" class="b">')

        self.elm.id = 'c'
        self.assertIn('src="a"', self.elm.start_tag)
        self.assertIn('id="c"', self.elm.start_tag)

    def test_inner_html(self):
        self.assertEqual(self.elm.innerHTML, '')
        self.elm.appendChild(Element('a'))
        self.assertEqual(self.elm.innerHTML, '<a></a>')

        self.elm.innerHTML = '<b></b>'
        self.assertEqual(self.elm.innerHTML, '<b></b>')
        self.assertEqual(self.elm.firstChild.tag, 'b')
        self.assertTrue(isinstance(self.elm.firstChild, HTMLElement))

    def test_inner_html_nest(self):
        html = '<b><c>d</c>e</b>'
        self.elm.innerHTML = html
        self.assertEqual(self.elm.innerHTML, html)
        self.assertEqual(self.elm.firstChild.html, html)
        self.assertEqual(self.elm.firstChild.firstChild.html, '<c>d</c>')
        self.assertEqual(self.elm.firstChild.firstChild.innerHTML, 'd')
        self.assertEqual(self.elm.firstChild.lastChild.html, 'e')
        self.assertTrue(isinstance(self.elm.firstChild.lastChild, Text))

    def test_parse_html_text(self):
        html = '''
        <a>a1</a1>
        b
        '''
        self.elm.innerHTML = html
        # fisrt node is empty (\n+spaces) text node
        self.assertEqual(self.elm.childNodes.length, 3)
        self.assertTrue(isinstance(self.elm.firstChild, Text))
        self.assertTrue(isinstance(self.elm.lastChild, Text))
        self.assertTrue(isinstance(self.elm.firstElementChild, HTMLElement))
        self.assertTrue(isinstance(self.elm.lastElementChild, HTMLElement))

    def test_insert_adjacent_html(self):
        self.elm.appendChild(self.c1)
        self.c1.insertAdjacentHTML('beforebegin', '<a></a>')
        self.assertEqual(self.elm.childNodes.length, 2)
        self.assertIs(self.elm.lastElementChild, self.c1)
        self.c1.insertAdjacentHTML('afterend', 'text')
        self.assertEqual(self.elm.childNodes.length, 3)
        self.assertIs(self.elm.lastElementChild, self.c1)
        self.c1.insertAdjacentHTML('afterBegin', '<b></b>')
        self.assertEqual(self.c1.childNodes.length, 1)
        self.c1.insertAdjacentHTML('BeforeEnd', '<c></c>')
        self.assertEqual(self.c1.childNodes.length, 2)
        with self.assertRaises(ValueError):
            self.c1.insertAdjacentHTML('a', 'b')

    def test_end_tag(self):
        self.assertEqual(self.elm.end_tag, '</tag>')

    def test_html(self):
        self.assertEqual(self.elm.html, '<tag></tag>')

    def test_append_string(self):
        with self.assertRaises(TypeError):
            self.elm.appendChild('a')
        self.assertFalse(self.elm.hasChildNodes())

    def test_get_elements_by_tagname(self):
        self.elm.appendChild(self.c1)
        self.elm.appendChild(self.c2)
        c1_tags = self.elm.getElementsByTagName('c1')
        c2_tags = self.elm.getElementsByTagName('c2')
        self.assertEqual(len(c1_tags), 1)
        self.assertEqual(len(c2_tags), 1)
        self.assertIs(c1_tags[0], self.c1)
        self.assertIs(c2_tags[0], self.c2)

    def test_get_elements_by_tagname_nest(self):
        self.elm.appendChild(self.c1)
        self.c1.appendChild(self.c2)
        c2_tags = self.c1.getElementsByTagName('c2')
        self.assertEqual(len(c2_tags), 1)
        self.assertIs(c2_tags[0], self.c2)
        c2_tags = self.elm.getElementsByTagName('c2')
        self.assertEqual(len(c2_tags), 1)
        self.assertIs(c2_tags[0], self.c2)

    def test_get_elements_by_tagname_self(self):
        c1_tags = self.c1.getElementsByTagName('c1')
        self.assertEqual(len(c1_tags), 0)

    def test_get_elements_by_classname(self):
        self.elm.appendChild(self.c1)
        self.elm.appendChild(self.c2)
        c1_classes = self.elm.getElementsByClassName('c1')
        c2_classes = self.elm.getElementsByClassName('c2')
        self.assertEqual(len(c1_classes), 1)
        self.assertEqual(len(c2_classes), 1)
        self.assertIs(c1_classes[0], self.c1)
        self.assertIs(c2_classes[0], self.c2)

    def test_get_elements_by_classname_nest(self):
        self.elm.appendChild(self.c1)
        self.c1.appendChild(self.c2)
        c2_classes = self.c1.getElementsByClassName('c2')
        self.assertEqual(len(c2_classes), 1)
        self.assertIs(c2_classes[0], self.c2)
        c2_classes = self.elm.getElementsByClassName('c2')
        self.assertEqual(len(c2_classes), 1)
        self.assertIs(c2_classes[0], self.c2)

    def test_clone_shallow_child(self):
        self.elm.appendChild(self.c1)
        clone = self.elm.cloneNode()
        self.assertFalse(clone.hasChildNodes())

        clone.appendChild(self.c2)
        self.assertFalse(self.c2 in self.elm)
        self.assertEqual(len(self.elm.childNodes), 1)

    def test_clone_shallow_attr(self):
        self.elm.setAttribute('src', 'a')
        clone = self.elm.cloneNode()

        self.assertTrue(clone.hasAttribute('src'))
        self.assertTrue(clone.getAttribute('src'), 'a')
        self.assertEqual(self.elm.attributes.toString(),
                         clone.attributes.toString())
        clone.setAttribute('src', 'b')
        self.assertNotEqual(self.elm.attributes.toString(),
                            clone.attributes.toString())
        self.assertTrue(self.elm.getAttribute('src'), 'a')
        self.assertTrue(clone.getAttribute('src'), 'b')

    def test_clone_deep_child(self):
        self.elm.appendChild(self.c1)
        clone = self.elm.cloneNode(deep=True)
        self.assertTrue(clone.hasChildNodes())
        self.assertEqual(len(clone.childNodes), 1)
        self.assertTrue(self.c1 in self.elm)
        self.assertFalse(self.c1 in clone)
        self.assertIsNot(self.c1, clone.firstChild)

        clone.appendChild(self.c2)
        self.assertFalse(self.c2 in self.elm)
        self.assertEqual(len(self.elm.childNodes), 1)
        self.assertEqual(len(clone.childNodes), 2)

    def test_clone_deep_attr(self):
        self.elm.setAttribute('src', 'a')
        self.elm.appendChild(self.c1)
        self.c1.setAttribute('src', 'c1')
        clone = self.elm.cloneNode(deep=True)

        self.assertTrue(clone.hasAttribute('src'))
        self.assertTrue(clone.getAttribute('src'), 'a')
        self.assertTrue(clone.firstChild.hasAttribute('src'))
        self.assertTrue(clone.firstChild.getAttribute('src'), 'c1')
        self.assertEqual(self.elm.attributes.toString(),
                         clone.attributes.toString())
        self.assertEqual(self.c1.attributes.toString(),
                         clone.firstChild.attributes.toString())

        clone.firstChild.setAttribute('src', 'b')
        self.assertNotEqual(self.c1.attributes.toString(),
                            clone.firstChild.attributes.toString())
        self.assertTrue(self.c1.getAttribute('src'), 'a')
        self.assertTrue(clone.firstChild.getAttribute('src'), 'b')

    def test_init_class(self):
        elm = Element('a', class_='a')
        self.assertEqual(elm.html, '<a class="a"></a>')
        self.assertEqual(elm.classList.length, 1)
        self.assertIn('a', elm.classList)

        elm2 = Element('a', **{'class': 'b'})
        self.assertEqual(elm2.html, '<a class="b"></a>')
        self.assertEqual(elm2.classList.length, 1)
        self.assertIn('b', elm2.classList)

    def test_init_class_multi_str(self):
        elm = Element('a', class_='a1 a2')
        self.assertEqual(elm.html, '<a class="a1 a2"></a>')
        self.assertEqual(elm.classList.length, 2)
        self.assertIn('a1', elm.classList)
        self.assertIn('a2', elm.classList)
        self.assertNotIn('a1 a2', elm.classList)

    def test_init_class_multi_list(self):
        elm = Element('a', class_=['a1', 'a2'])
        self.assertEqual(elm.html, '<a class="a1 a2"></a>')
        self.assertEqual(elm.classList.length, 2)
        self.assertIn('a1', elm.classList)
        self.assertIn('a2', elm.classList)
        self.assertNotIn('a1 a2', elm.classList)

    def test_init_class_multi_mixed(self):
        elm = Element('a', class_=['a1', 'a2 a3'])
        self.assertEqual(elm.html, '<a class="a1 a2 a3"></a>')
        self.assertEqual(elm.classList.length, 3)
        self.assertIn('a1', elm.classList)
        self.assertIn('a2', elm.classList)
        self.assertIn('a3', elm.classList)
        self.assertNotIn('a2 a3', elm.classList)

    def test_reference(self):
        gc.collect()
        elm = Element('a')
        _id = id(elm)
        self.assertIn(elm, Element._elements)
        del elm
        gc.collect()  # run gc
        for elm in Element._elements:
            assert id(elm) != _id

    def test_reference_with_id(self):
        gc.collect()
        elm = Element('a', id='a')
        _id = id(elm)
        self.assertIn(elm.id, Element._elements_with_id)
        del elm
        gc.collect()
        self.assertNotIn('a', Element._elements_with_id)
        for elm in Element._elements:
            assert id(elm) != _id

    def test_reference_add_id(self):
        gc.collect()
        elm = Element('a')
        _id = id(elm)
        self.assertNotIn(elm, Element._elements_with_id.values())
        elm.id = 'a'
        self.assertIn('a', Element._elements_with_id)
        self.assertIn(elm, Element._elements_with_id.values())
        elm.id = 'b'
        self.assertNotIn('a', Element._elements_with_id)
        self.assertIn('b', Element._elements_with_id)
        self.assertIn(elm, Element._elements_with_id.values())
        elm.setAttribute('id', 'c')
        self.assertNotIn('b', Element._elements_with_id)
        self.assertIn('c', Element._elements_with_id)
        self.assertIn(elm, Element._elements_with_id.values())
        del elm
        gc.collect()
        self.assertNotIn('c', Element._elements_with_id)
        for elm in Element._elements:
            assert id(elm) != _id

    def test_reference_del_id(self):
        gc.collect()
        elm = Element('a', id='a')
        self.assertIn('a', Element._elements_with_id)
        self.assertIn(elm, Element._elements_with_id.values())
        elm.removeAttribute('id')
        self.assertNotIn('a', Element._elements_with_id)
        self.assertNotIn(elm, Element._elements_with_id.values())

    def test_is_attr(self):
        '''``is`` is a reserved word for python, so use ``is_`` in constructor.
        '''
        elm = Element('tag', is_='elm')
        self.assertIn('is', elm.attributes)
        self.assertNotIn('is_', elm.attributes)
        self.assertEqual('elm', elm.getAttribute('is'))
        self.assertIsNone(elm.getAttribute('is_'))

        # ``is_`` is not treated as special at setAttribute
        elm.setAttribute('is_', 'new')
        self.assertEqual('elm', elm.getAttribute('is'))
        self.assertEqual('new', elm.getAttribute('is_'))

    class NewTag(HTMLElement):
        pass

    def test_custom_tag(self):
        self.elm.innerHTML = '<new-tag></new-tag>'
        child = self.elm.firstChild
        self.assertEqual(child.__class__, HTMLElement)
        customElements.define('new-tag', self.NewTag)
        self.assertEqual(child.__class__, self.NewTag)

    def test_custom_tag_registered(self):
        customElements.define('new-tag', self.NewTag)
        self.elm.innerHTML = '<new-tag></new-tag>'
        self.assertEqual(self.elm.firstChild.__class__, self.NewTag)

    def test_custom_tag_is(self):
        self.elm.innerHTML = '<a is="my-a"></a>'
        child = self.elm.firstChild
        self.assertEqual(child.__class__, HTMLElement)
        self.assertEqual(child.getAttribute('is'), 'my-a')
        customElements.define('my-a', self.NewTag, {'extends': 'a'})
        self.assertEqual(self.elm.firstChild.__class__, self.NewTag)

    def test_custom_tag_is_registered(self):
        customElements.define('my-a', self.NewTag, {'extends': 'a'})
        self.elm.innerHTML = '<a is="my-a"></a>'
        self.assertEqual(self.elm.firstChild.__class__, self.NewTag)

    def test_invalid_define_args(self):
        with self.assertRaises(TypeError):
            customElements.define(1, 2, 3)


class TestHTMLElement(TestCase):
    def setUp(self):
        self.elm = HTMLElement('a')

    def test_attrs_bool(self):
        self.assertFalse(self.elm.hasAttribute('hidden'))
        self.assertNotIn('hidden', self.elm.attributes)
        self.assertIsNone(self.elm.getAttribute('hidden'))
        self.elm.setAttribute('hidden', True)

        self.assertTrue(self.elm.hasAttributes())
        self.assertTrue(self.elm.hasAttribute('hidden'))
        self.assertIn('hidden', self.elm.attributes)
        self.assertEqual(self.elm.getAttribute('hidden'), True)
        self.assertEqual(self.elm.html, '<a hidden></a>')

        # This is complatible with JS, but quite confusing
        self.elm.setAttribute('hidden', False)
        self.assertTrue(self.elm.hasAttributes())
        self.assertTrue(self.elm.hasAttribute('hidden'))
        self.assertIn('hidden', self.elm.attributes)
        # In chrome, <a hidden="false">, should we convert to string?
        self.assertFalse(self.elm.getAttribute('hidden'))
        self.assertEqual(self.elm.html, '<a hidden></a>')

        self.elm.hidden = False
        self.assertFalse(self.elm.hasAttribute('hidden'))
        self.assertNotIn('hidden', self.elm.attributes)
        self.assertIsNone(self.elm.getAttribute('hidden'))
        self.assertEqual(self.elm.html, '<a></a>')

    def test_empty_tag(self):
        img = HTMLElement('img')
        self.assertEqual(img.end_tag, '')

    def test_draggable(self):
        n = HTMLElement('img')
        n.draggable = True
        self.assertEqual(n.start_tag, '<img draggable>')

    def test_hidden(self):
        n = HTMLElement('img')
        n.hidden = True
        self.assertEqual(n.start_tag, '<img hidden>')

    def test_title(self):
        n = HTMLElement('img')
        n.title = 'Image'
        self.assertEqual(n.start_tag, '<img title="Image">')

    def test_type(self):
        n = HTMLElement('input')
        n.type = 'text'
        self.assertEqual(n.start_tag, '<input type="text">')

    def test_init_attrs(self):
        elm = HTMLElement('a', src='b', hidden=True)
        self.assertFalse(elm.hasChildNodes())
        self.assertTrue(elm.hasAttributes())
        self.assertTrue(elm.getAttribute('src'), 'b')
        self.assertTrue(elm.hidden)

    def test_style_empty(self):
        self.assertIsNotNone(self.elm.style)
        self.assertEqual(self.elm.getAttribute('style'), None)
        self.assertEqual(self.elm.style.cssText, '')
        self.assertTrue(isinstance(self.elm.style, CSSStyleDeclaration))

        self.assertEqual(self.elm.html, '<a></a>')

    def test_style_invalid_type(self):
        with self.assertRaises(TypeError):
            self.elm.style = 1
        with self.assertRaises(TypeError):
            self.elm.style = self.elm

    def test_init_style_init(self):
        elm = HTMLElement('a', style='color: red;')
        self.assertEqual(elm.style.cssText, 'color: red;')
        self.assertEqual(elm.getAttribute('style'), 'color: red;')
        self.assertEqual(elm.html, '<a style="color: red;"></a>')

    @skipIf(sys.version_info < (3, 5), 'py34 does not keep style order')
    def test_style_setter(self):
        self.elm.style = 'color: red;'
        self.assertEqual(self.elm.style.cssText, 'color: red;')
        self.assertEqual(self.elm.getAttribute('style'), 'color: red;')
        self.assertEqual(self.elm.html, '<a style="color: red;"></a>')

        self.elm.style.color = 'black'
        self.elm.style.background = 'red'
        # py34 does not keep style order
        self.assertEqual(self.elm.style.cssText,
                        'color: black; background: red;')

    def test_style_remove(self):
        self.elm.style = 'color: red;'
        self.elm.removeAttribute('style')
        self.assertIsNotNone(self.elm.style)
        self.assertEqual(self.elm.getAttribute('style'), None)
        self.assertEqual(self.elm.style.cssText, '')

        self.assertEqual(self.elm.html, '<a></a>')

    def test_style_clone(self):
        self.elm.style = 'color: red;'
        clone = self.elm.cloneNode()
        self.assertEqual(clone.style.cssText, 'color: red;')

        clone.style.color = 'black'
        self.assertEqual(clone.style.cssText, 'color: black;')
        self.assertEqual(self.elm.style.cssText, 'color: red;')

    def test_attr_clone(self):
        self.elm.draggable = True
        self.elm.hidden = True
        clone = self.elm.cloneNode()
        self.assertEqual(clone.html, '<a draggable hidden></a>')
        self.elm.hidden = False
        self.assertEqual(clone.html, '<a draggable hidden></a>')
        clone.draggable = False
        self.assertEqual(self.elm.html, '<a draggable></a>')
        self.assertEqual(clone.html, '<a hidden></a>')

    def test_attr_clone_deep(self):
        self.elm.draggable = True
        self.elm.hidden = True
        clone = self.elm.cloneNode(deep=True)
        self.assertEqual(clone.html, '<a draggable hidden></a>')
        self.elm.hidden = False
        self.assertEqual(clone.html, '<a draggable hidden></a>')
        clone.draggable = False
        self.assertEqual(self.elm.html, '<a draggable></a>')
        self.assertEqual(clone.html, '<a hidden></a>')


class TestSelectElement(TestCase):
    def setUp(self):
        self.select = HTMLSelectElement('select')
        self.opt1 = HTMLOptionElement('option')
        self.opt2 = HTMLOptionElement('option')
        self.opt3 = HTMLOptionElement('option')

    def test_options(self) -> None:
        self.assertEqual(self.select.length, 0)
        self.assertEqual(self.select.options.length, 0)
        self.select.append(self.opt1)
        self.assertEqual(self.select.length, 1)
        self.assertEqual(self.select.options.length, 1)
        self.select.append(self.opt2, self.opt3)
        self.assertEqual(self.select.length, 3)
        self.assertEqual(self.select.options.length, 3)

    def test_selected(self) -> None:
        self.select.append(self.opt1, self.opt2, self.opt3)
        for opt in self.select.options:
            self.assertFalse(opt.selected)
        self.assertEqual(self.select.selectedOptions.length, 0)
        self.opt1.selected = True
        self.assertEqual(self.select.selectedOptions.length, 1)
        self.opt2.selected = True
        self.assertEqual(self.select.selectedOptions.length, 2)
        self.opt3.selected = True
        self.assertEqual(self.select.selectedOptions.length, 3)
        self.opt1.selected = False
        self.assertEqual(self.select.selectedOptions.length, 2)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
import asyncio
import socket
from typing import Optional

from aiohttp import web, MsgType

from wdom.options import config
from wdom.server.handler import on_websocket_message

logger = logging.getLogger(__name__)
connections = []
_access_log_format = '%s %r (%a) %D us'


def is_connected():
    """Check if aiohttp server has a client connection."""
    return any(connections)


def main_handler(request):
    """Main handler to serve root ``document`` object of the application."""
    from wdom.document import get_document
    logger.info('connected')
    return web.Response(body=get_document().build().encode())


@asyncio.coroutine
def ws_open(request):
    """Open websocket connection of aiohttp."""
    handler = WSHandler()
    logger.info('WS OPEN')
    yield from handler.open(request)
    return handler.ws


class WSHandler:
    """Wrapper class for aiohttp websockets.

    APIs are similar to ``tornaod.websocket.WebSocketHandler``.
    """
    @asyncio.coroutine
    def open(self, request):
        """Open new websocket connection and keep it."""
        self.req = request
        self.ws = web.WebSocketResponse()
        yield from self.ws.prepare(request)
        connections.append(self)

        while not self.ws.closed:
            msg = yield from self.ws.receive()
            if msg.tp == MsgType.text:
                yield from self.on_message(msg.data)
            elif msg.tp in (MsgType.close, MsgType.closed, MsgType.error):
                yield from self.ws.close()
        self.on_close()
        return self.ws

    def write_message(self, message):
        """Send message to the client."""
        self.ws.send_str(message)

    @asyncio.coroutine
    def on_message(self, message):
        """Called when get message from client."""
        on_websocket_message(message)

    @asyncio.coroutine
    def terminate(self):
        """Terminate if no more connection."""
        yield from asyncio.sleep(config.shutdown_wait)
        # stop server and close loop if no more connection exists
        if not is_connected():
            server = self.req.app['server']
            yield from terminate_server(server)
            server._loop.stop()

    def on_close(self):
        """Called when connection closed."""
        logger.info('RootWS CLOSED')
        if self in connections:
            # Remove this connection from connection-list
            connections.remove(self)
        # close if auto_shutdown is enabled and there is no more connection
        if config.auto_shutdown and not is_connected():
            asyncio.ensure_future(self.terminate())


class Application(web.Application):
    """Web server application class."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # self.router.add_route('GET', '/', MainHandler)
        root_resource = self.router.add_resource('/', name='root')
        root_resource.add_route('GET', main_handler)
        # self.router.add_route('*', '/rimo_ws', ws_open)
        root_ws_resource = self.router.add_resource('/rimo_ws', name='root_ws')
        root_ws_resource.add_route('*', ws_open)

    def add_static_path(self, prefix: str, path: str):
        """Add directory to serve static files."""
        if not prefix.startswith('/'):
            prefix = '/' + prefix
        self.router.add_static(prefix, path)

    def add_favicon_path(self, path: str):
        """Add path to serve favicon file."""
        self.router.add_static('/(favicon.ico)', path)


main_application = Application()


def get_app(*args, **kwargs) -> web.Application:
    """Make Application object to serve ``document``."""
    return main_application


def set_application(app: Application):
    """Set application as a root application."""
    global main_application
    main_application = app


@asyncio.coroutine
def close_connections(app: web.Application):
    """Close all websocket connections."""
    for conn in connections:
        yield from conn.ws.close(code=999, message='server shutdown')


def start_server(app: Optional[web.Application] = None,
                 port: Optional[int] = None,
                 browser: Optional[str] = None,
                 loop: Optional[asyncio.BaseEventLoop] = None,
                 address: Optional[str] = None,
                 family: Optional[socket.AddressFamily] = socket.AF_INET,
                 check_time: Optional[int] = 500,
                 ) -> asyncio.base_events.Server:
    """Start server with ``app`` on ``address:port``.

    If port is not specified, use command line option of ``--port``.

    When ``browser`` is specified, open the page with the specified browser.
    The specified browser name is not registered in ``webbrowser`` module, or,
    for example it is just ``True``, use system's default browser to open the
    page.
    """
    port = port if port is not None else config.port
    address = address if address is not None else config.address
    app = app or get_app()
    loop = loop or asyncio.get_event_loop()

    handler = app.make_handler(logger=logger, access_log=logger,
                               access_log_format=_access_log_format)
    f = loop.create_server(handler, address, port)
    server = loop.run_until_complete(f)
    server.app = app
    server.handler = handler
    server.port = server.sockets[-1].getsockname()[1]
    server.address = address or 'localhost'
    app.on_shutdown.append(close_connections)
    app['server'] = server

    return server


@asyncio.coroutine
def terminate_server(server: asyncio.base_events.Server):
    """Close all connections and terminate server."""
    logger.info('Start server shutdown')
    server.close()
    yield from server.wait_closed()
    yield from server.app.shutdown()
    yield from server.handler.finish_connections(1.0)
    yield from server.app.cleanup()
    logger.info('Server terminated')


def stop_server(server: asyncio.base_events.Server):
    """Terminate given server."""
    server._loop.run_until_complete(terminate_server(server))
